/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "nntpgrab_automation.h"
#include "nntpgrab_utils.h"
#include <glib/gi18n.h>

#if GLIB_CHECK_VERSION(2,16,0)
#include <gio/gio.h>

static GFileMonitor *monitor = NULL;
static GHashTable *queued_files = NULL;

struct _queued_file {
    time_t stamp_last_seen;
};

static gboolean process_queue(gpointer data);

static void
directory_monitor_changed (GFileMonitor *monitor, GFile *file, GFile *other_file, GFileMonitorEvent event_type, gpointer user_data)
{
    char *filename;
    struct _queued_file *queued_file;

    if (event_type != G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT &&
        event_type != G_FILE_MONITOR_EVENT_CREATED &&
        event_type != G_FILE_MONITOR_EVENT_CHANGED &&
        queued_files != NULL) {

        return;
    }

    // only handle .nzb files
    filename = g_file_get_path(file);

    g_return_if_fail(filename != NULL);

    if (strlen(filename) < 3 || g_strncasecmp(filename + strlen(filename) - 4, ".nzb", 3)) {
        g_print("File '%s' detected by the NZB auto import, but ignored as it has the wrong extension\n", filename);
        g_free(filename);
        return;
    }

    // Have we already seen this file recently?
    queued_file = g_hash_table_lookup(queued_files, filename);
    if (!queued_file) {
        queued_file = g_slice_alloc0(sizeof(struct _queued_file));
        g_hash_table_insert(queued_files, filename, queued_file);

        if (g_hash_table_size(queued_files) == 1) {
            g_timeout_add_seconds(1, process_queue, user_data);
        }
    }

    queued_file->stamp_last_seen = time(NULL);
}

static void
process_file(const char *filename, gpointer user_data)
{
    char *contents = NULL;
    GError *err = NULL;
    NZB *nzbfile;
    char *errmsg = NULL;
    ConfigOpts opts;
    char *filename_stripped;

    if (!g_file_get_contents(filename, &contents, NULL, &err)) {
        g_warning(__FILE__ ":%i Error occured while reading file\n%s:\n%s", __LINE__, filename, err->message);
        g_error_free(err);
        return;
    }

    nzbfile = nntpgrab_utils_parse_nzb_file(contents, &errmsg);
    if (!nzbfile) {
        if (!errmsg) {
            errmsg = g_strdup_printf(_("Error occured while parsing NZB File '%s'\n"), filename);
        }

        g_warning(errmsg);

        g_free(errmsg);
        g_free(contents);

        return;
    }

    filename_stripped = g_path_get_basename(filename);
    nntpgrab_utils_strip_nzb_extension(filename_stripped);

    nntpgrab_automation_import_nzb_file((NNTPGrabCore*) user_data, nzbfile, filename_stripped, &errmsg);
    if (errmsg) {
        g_warning(errmsg);
        g_free(errmsg);
    }
    g_free(filename_stripped);
    g_free(contents);

    opts = nntpgrab_config_get_opts((NNTPGrabCore*) user_data);
    if (opts.move_file_after_auto_import) {
        char *basename_orig;
        char *dirname_orig;
        char *dirname;
        char *path;

        dirname_orig = g_path_get_dirname(filename);
        dirname = g_build_filename(dirname_orig, "Imported", NULL);
        g_mkdir_with_parents(dirname, 0700);

        basename_orig = g_path_get_basename(filename);
        path = g_build_filename(dirname, basename_orig, NULL);

        g_rename(filename, path);

        g_free(path);
        g_free(dirname);
        g_free(dirname_orig);
        g_free(basename_orig);
    } else {
        g_unlink(filename);
    }
}

static gboolean
process_queue(gpointer data)
{
    GHashTableIter iter;
    gpointer key, value;
    static int times_been_here_without_processing = 0;
    gboolean processed_something = FALSE;

    if (!queued_files || times_been_here_without_processing > 10) {
        // We need to stop monitoring
        times_been_here_without_processing = 0;
        return FALSE;
    }

    g_hash_table_iter_init (&iter, queued_files);
    while (g_hash_table_iter_next (&iter, &key, &value)) {
        struct _queued_file *queued_file = (struct _queued_file *) value;

        // Only process files which haven't changed in the last 3 seconds
        if (queued_file->stamp_last_seen + 3 > time(NULL)) {
            continue;
        }

        process_file((const char *) key, data);
        g_hash_table_iter_remove(&iter);
        processed_something = TRUE;
    }

    if (processed_something) {
        times_been_here_without_processing = 0;
    } else {
        times_been_here_without_processing++;
    }

    return TRUE;
}

static void
queued_file_free(gpointer data)
{
    g_slice_free(struct _queued_file, data);
}

gboolean
nntpgrab_automation_enable_auto_import(NNTPGrabCore *core, const char *directory, char **errmsg)
{
    if (monitor) {
        nntpgrab_utils_cancel_directory_monitor(monitor);
    }

    queued_files = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, queued_file_free);

    monitor = nntpgrab_utils_monitor_directory(directory, errmsg);
    if (!monitor) {
        return FALSE;
    }

    g_signal_connect(monitor, "changed", G_CALLBACK(directory_monitor_changed), core);

    return TRUE;
}

void
nntpgrab_automation_disable_auto_import(void)
{
    if (monitor) {
        nntpgrab_utils_cancel_directory_monitor(monitor);
        monitor = NULL;
        g_hash_table_unref(queued_files);
        queued_files = NULL;
    }
}

#endif
