/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include "gui.h"

static GHashTable *active_connections = NULL;

static GtkTreeRowReference *
get_active_connection_row_ref(const char *servername, int conn_id)
{
    GtkTreeRowReference *ref;

    ref = g_hash_table_lookup(active_connections, GINT_TO_POINTER(conn_id));
    if (!ref) {
        GtkWidget *treeConnections;
        GtkTreeModel *store;
        GtkTreeIter iter;
        GtkTreePath *path;

        treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
        store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

        // No element found, add a new one
        gtk_list_store_append(GTK_LIST_STORE(store), &iter);
        gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, conn_id, 1, servername, 2, _("Connecting"), -1);

        path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);

        ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(store), path);
        g_hash_table_insert(active_connections, GINT_TO_POINTER(conn_id), ref);

        gtk_tree_path_free(path);
    }

    return ref;
}

static void
connection_connecting (NNTPGrabCore *core, const char *servername, int conn_id, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    GtkTreePath *path;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    // Is this connection id already known?
    if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter)) {
        do {
            int id;

            gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, 0, &id, -1);

            if (id == conn_id) {
                gtk_list_store_set(GTK_LIST_STORE(store), &iter, 1, servername, 2, _("Connecting"), 4, "", -1);
                return;
            }
        } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &iter));
    }

    // No element found, add a new one
    gtk_list_store_append(GTK_LIST_STORE(store), &iter);
    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, conn_id, 1, servername, 2, _("Connecting"), -1);

    path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);

    g_hash_table_insert(active_connections, GINT_TO_POINTER(conn_id), gtk_tree_row_reference_new(GTK_TREE_MODEL(store), path));

    gtk_tree_path_free(path);
}

static void
connection_connected (NNTPGrabCore *core, const char *servername, int conn_id, const char *welcome_msg, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    GtkTreeRowReference *ref;
    GtkTreePath *path;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    ref = get_active_connection_row_ref(servername, conn_id);

    path = gtk_tree_row_reference_get_path(ref);
    gtk_tree_model_get_iter(store, &iter, path);
    gtk_tree_path_free(path);

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Connected"), 4, welcome_msg, -1);
}

static gboolean
remove_idle_connection(gpointer data)
{
    GtkTreeRowReference *ref = (GtkTreeRowReference *) data;

    if (gtk_tree_row_reference_valid(ref)) {
        GtkTreeModel *model = gtk_tree_row_reference_get_model(ref);
        GtkTreePath *path;
        GtkTreeIter iter;
        char *label = NULL;
        int conn_id;

        path = gtk_tree_row_reference_get_path(ref);
        gtk_tree_model_get_iter(model, &iter, path);
        gtk_tree_path_free(path);

        gtk_tree_model_get(model, &iter, 0, &conn_id, 2, &label, -1);

        if (!strcmp(label, _("Disconnected"))) {
            // Connection is still inactive, remove it
            gtk_list_store_remove(GTK_LIST_STORE(model), &iter);

            g_hash_table_remove(active_connections, GINT_TO_POINTER(conn_id));
        }

        g_free(label);
    }

    gtk_tree_row_reference_free(ref);

    return FALSE;
}

static void
connection_disconnect (NNTPGrabCore *core, const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *disconnect_msg, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    GtkTreeRowReference *ref;
    GtkTreePath *path;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    ref = get_active_connection_row_ref(servername, conn_id);

    path = gtk_tree_row_reference_get_path(ref);
    gtk_tree_model_get_iter(store, &iter, path);

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Disconnected"), 4, disconnect_msg, -1);

    // Add a timer so this record will automatically be removed after 10 seconds of inactivity
    path = gtk_tree_model_get_path(store, &iter);
    ref = gtk_tree_row_reference_new(store, path);
    gtk_tree_path_free(path);

    g_timeout_add(1000 * 10, remove_idle_connection, ref);
}

static void
part_download_start (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    GtkTreeRowReference *ref;
    GtkTreePath *path;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    ref = get_active_connection_row_ref(servername, conn_id);

    path = gtk_tree_row_reference_get_path(ref);
    gtk_tree_model_get_iter(store, &iter, path);
    gtk_tree_path_free(path);

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Downloading"), 4, subject, -1);
}

static void
part_progress_update (NNTPGrabCore *core, const char *servername, int conn_id, const char *subject, int part_num, int bytes_downloaded, int bytes_total, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    GtkTreeRowReference *ref;
    GtkTreePath *path;
    double progress;
    int progress_int;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    ref = get_active_connection_row_ref(servername, conn_id);

    path = gtk_tree_row_reference_get_path(ref);
    gtk_tree_model_get_iter(store, &iter, path);
    gtk_tree_path_free(path);

    progress = bytes_downloaded;
    progress /= bytes_total;
    progress *= 100;
    progress_int = (int) progress;

    if (progress_int > 100) {
        g_print(__FILE__ ":%i progress overflow!\n", __LINE__);
        progress_int = 100;
    }

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Downloading"), 3, progress_int, -1);
}

static void
part_download_done (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    GtkTreeRowReference *ref;
    GtkTreePath *path;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    ref = get_active_connection_row_ref(servername, conn_id);

    path = gtk_tree_row_reference_get_path(ref);
    gtk_tree_model_get_iter(store, &iter, path);
    gtk_tree_path_free(path);

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Idle"), 3, 0, 4, "", -1);
}

void
connections_initialize(void)
{
    GtkWidget *treeConnections;
    GtkListStore *store;
    GtkCellRenderer *cellRenderer;

    active_connections = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, (GDestroyNotify) gtk_tree_row_reference_free);

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConnections), -1, _("Servername"), cellRenderer, "text", 1, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConnections), -1, _("Status"), cellRenderer, "text", 2, NULL);

    cellRenderer = gtk_cell_renderer_progress_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConnections), -1, _("Part progress"), cellRenderer, "value", 3, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConnections), -1, _("Filename"), cellRenderer, "text", 4, NULL);

    store = gtk_list_store_new(5, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);

    gtk_tree_view_set_model(GTK_TREE_VIEW(treeConnections), GTK_TREE_MODEL(store));

    g_signal_connect(core, "connection_connecting", G_CALLBACK(connection_connecting), NULL);
    g_signal_connect(core, "connection_connected", G_CALLBACK(connection_connected), NULL);
    g_signal_connect(core, "connection_disconnect", G_CALLBACK(connection_disconnect), NULL);
    g_signal_connect(core, "part_download_start", G_CALLBACK(part_download_start), NULL);
    g_signal_connect(core, "part_progress_update", G_CALLBACK(part_progress_update), NULL);
    g_signal_connect(core, "part_done", G_CALLBACK(part_download_done), NULL);
    g_signal_connect(core, "part_failed", G_CALLBACK(part_download_done), NULL);
}
