/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "gui.h"
#include <string.h>
#include <gtkhtml/gtkhtml.h>
#include <curl.h>
#include <types.h>
#include <easy.h>
#include <libxml/tree.h>
#ifdef WIN32
#include <gdk/gdkwin32.h>
#endif

static size_t
curl_write_function(void *buffer, size_t size, size_t nmemb, void *stream)
{
    char *buf;
    GString *data = (GString *) stream;

    g_assert(data);

    /* The buffer isn't \0 terminated. Solve this by
       using a temporary buffer */
    buf = (char *) g_slice_alloc((int) (size * nmemb) + 1);
    memcpy(buf, buffer, size * nmemb);
    buf[size * nmemb] = '\0';
    data = g_string_append(data, buf);
    g_slice_free1((size * nmemb) + 1, buf);

    return (int) (size * nmemb);
}

static gboolean
open_uri (GtkWidget *embed, const char *url_unquoted, gpointer data)
{
    if (strstr(url_unquoted, "http://")) {
        char *url = g_shell_quote(url_unquoted);
#ifdef WIN32
        GtkWidget *windowMain = nntpgrab_gui_base_get_widget("windowMain");
        ShellExecuteA(GDK_WINDOW_HWND(windowMain->window), "open", url_unquoted, NULL, NULL, SW_SHOW);
#elif defined(DARWIN)
        GError *error = NULL;
        char cmd[1024];

        snprintf(cmd, sizeof(cmd), "open %s", url);
        g_spawn_command_line_async(cmd, &error);
        if (error) {
            g_print("%s: %s\n", error->message, cmd);
        }
#else
        GError *error = NULL;
        char cmd[1024];

        if (g_find_program_in_path("xdg-open")) {
            // We hebben de desktop-onafhankelijke methode om programma's te openen
            snprintf(cmd, sizeof(cmd), "xdg-open %s", url);
        } else if (g_getenv("GNOME_DESKTOP_SESSION_ID") && g_find_program_in_path("gnome-open")) {
            // We draaien Gnome
            snprintf(cmd, sizeof(cmd), "gnome-open %s", url);
        } else if (g_getenv("KDE_FULL_SESSION") && g_find_program_in_path("kfmclient")) {
            // We draaien KDE
            snprintf(cmd, sizeof(cmd), "kfmclient openURL %s", url);
        } else if (g_getenv("BROWSER")) {
            // XFCE4 gebruikt de environment variabele BROWSER om de standaardbrowser te bepalen.
            snprintf(cmd, sizeof(cmd), "%s %s", g_getenv("BROWSER"), url);
        } else {
            // Window manager onbekend..probeer firefox, mozilla of konqueror te lanceren indien aanwezig
            if (g_find_program_in_path("firefox")) {
                snprintf(cmd, sizeof(cmd), "firefox %s", url);
            } else if (g_find_program_in_path("mozilla")) {
                snprintf(cmd, sizeof(cmd), "mozilla %s", url);
            } else if (g_find_program_in_path("konqueror")) {
            snprintf(cmd, sizeof(cmd), "konqueror %s", url);
            } else {
                GtkWidget *windowMain = nntpgrab_gui_base_get_widget("windowMain");
                nntpgrab_gui_base_dialog_show(windowMain, _("Unable to find a external webbrowser"), GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);
                g_free(url);
                return FALSE;
            }
        }

        g_spawn_command_line_async(cmd, &error);
        if (error) {
            g_print("%s: %s\n", error->message, cmd);
        }
#endif

        g_free(url);

        return FALSE;
    }

    return TRUE;
}

#if GTK_CHECK_VERSION(2,12,0)
void
initialize_html_widget(void)
{
    GtkWidget *html = nntpgrab_gui_base_get_widget("htmlInformation");

    g_signal_connect(html, "link_clicked", G_CALLBACK(open_uri), NULL);
}
#else
GtkWidget *
html_information_widget_create(gchar *widget_name, gchar *string1, gchar *string2, gint int1, gint int2)
{
    GtkWidget *html = gtk_html_new();
    gtk_widget_show(html);

    g_signal_connect(html, "link_clicked", G_CALLBACK(open_uri), NULL);

    return html;
}
#endif

static char *
retrieve_website_info(void)
{
    CURL            *curl = NULL;
    CURLcode         curl_retval;
    GString         *serverOutput;
    char            *ret;
    char            *user_agent;

    curl_global_init(CURL_GLOBAL_WIN32);

    curl = curl_easy_init();
    if(!curl) {
        g_warning(_("retrieve_website_info(): curl_easy_init() failed\n"));
        return NULL;
    }

    serverOutput = g_string_new("");

    curl_easy_setopt(curl, CURLOPT_URL, "http://www.nntpgrab.nl/info.php");
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_write_function);
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, serverOutput);
    curl_easy_setopt(curl, CURLOPT_NOPROGRESS, TRUE);
    curl_easy_setopt(curl, CURLOPT_ENCODING, "gzip");
    curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 20);
    curl_easy_setopt(curl, CURLOPT_NOSIGNAL, TRUE);
    //curl_easy_setopt(curl, CURLOPT_VERBOSE, TRUE);

    user_agent = g_strdup("NNTPGrab version " VERSION " (" OS ")");
    curl_easy_setopt(curl, CURLOPT_USERAGENT, user_agent);

#if 0
    if (core->config->proxyHost && core->config->proxyPort) {
        curl_easy_setopt(curl, CURLOPT_PROXY, ftd_string_value(core->config->proxyHost));
        curl_easy_setopt(curl, CURLOPT_PROXYPORT, atoi(ftd_string_value(core->config->proxyPort)));

        if (core->config->proxyUser && core->config->proxyPassword) {
            int len = ftd_string_length(core->config->proxyUser) + ftd_string_length(core->config->proxyPassword) + 2;
            proxyUserPw = g_malloc (len);
            snprintf(proxyUserPw, len, "%s:%s", ftd_string_value(core->config->proxyUser), ftd_string_value(core->config->proxyPassword));
            curl_easy_setopt(curl, CURLOPT_PROXYUSERPWD, proxyUserPw);
        }
    }
#endif

    curl_retval = curl_easy_perform(curl);
    if (curl_retval == 0) {
        ret = g_strdup(serverOutput->str);
    } else {
        ret = NULL;
    }

    g_free(user_agent);
    g_string_free(serverOutput, TRUE);

    curl_easy_cleanup(curl);
    curl_global_cleanup();

    return ret;
}

typedef struct _changelog
{
    int major_version;
    int minor_version;
    int micro_version;
    GList *entries;
} ChangeLog;

typedef struct _update_info
{
    int stable_major_version;
    int stable_minor_version;
    int stable_micro_version;
    time_t date_latest_stable_version;
    char date_latest_stable_version_str[128];

    int unstable_major_version;
    int unstable_minor_version;
    int unstable_micro_version;
    time_t date_latest_unstable_version;
    char date_latest_unstable_version_str[128];

    GList *changelogs;
} UpdateInfo;

static char *xml_parser_errmsg = NULL;

static void
xml_parser_error_func (void *ctx, const char *msg, ...)
{
    va_list args;
    char *orig = xml_parser_errmsg;
    char *tmp;

    va_start(args, msg);

    tmp = g_strdup_vprintf(msg, args);
    if (orig) {
        xml_parser_errmsg = g_strdup_printf("%s %s", orig, tmp);
        g_free(orig);
        g_free(tmp);
    } else {
        xml_parser_errmsg = tmp;
    }

    va_end(args);
}

static void
parse_changelog(xmlDocPtr doc, xmlNodePtr cur, ChangeLog *changelog)
{
    while(cur != NULL) {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "VersionMajor")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            changelog->major_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "VersionMinor")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            changelog->minor_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "VersionMicro")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            changelog->micro_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "Entry")) {
            xmlChar *entry = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            changelog->entries = g_list_append(changelog->entries, g_strdup((const char *) entry));
            xmlFree(entry);
        }

        cur = cur->next;
    }
}

static void
parse_changelogs(xmlDocPtr doc, xmlNodePtr cur, UpdateInfo *info)
{
    while(cur != NULL) {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "Version")) {
            ChangeLog *changelog = g_slice_new0(ChangeLog);

            info->changelogs = g_list_append(info->changelogs, changelog);

            parse_changelog(doc, cur->xmlChildrenNode, changelog);
        }

        cur = cur->next;
    }
}

static void
parse_info(xmlDocPtr doc, xmlNodePtr cur, UpdateInfo *info)
{
    while(cur != NULL) {
        // stable
        if (!xmlStrcmp(cur->name, (const xmlChar *) "LatestStableMajor")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            info->stable_major_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "LatestStableMinor")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            info->stable_minor_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "LatestStableMicro")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            info->stable_micro_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "DateLatestVersion")) {
            struct tm *now;

            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            info->date_latest_stable_version = atoi((const gchar *) version);
            xmlFree(version);

            if ((now = localtime(&info->date_latest_stable_version)) != NULL) {
                strftime(info->date_latest_stable_version_str, sizeof(info->date_latest_stable_version_str), "%x", now);
            }
        }

        // unstable
        if (!xmlStrcmp(cur->name, (const xmlChar *) "LatestUnstableMajor")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            info->unstable_major_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "LatestUnstableMinor")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            info->unstable_minor_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "LatestUnstableMicro")) {
            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            info->unstable_micro_version = atoi((const gchar *) version);
            xmlFree(version);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "DateLatestUnstable")) {
            struct tm *now;

            xmlChar *version = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            info->date_latest_unstable_version = atoi((const gchar *) version);
            xmlFree(version);

            if ((now = localtime(&info->date_latest_unstable_version)) != NULL) {
                strftime(info->date_latest_unstable_version_str, sizeof(info->date_latest_unstable_version_str), "%x", now);
            }
        }
        cur = cur->next;
    }
}

static UpdateInfo *
parse_version_info(const char *contents, char **errmsg)
{
    xmlDocPtr    doc;
    xmlNodePtr   cur;
    UpdateInfo  *info = NULL;

    xmlSetGenericErrorFunc(NULL, xml_parser_error_func);

    if (xml_parser_errmsg) {
        g_free(xml_parser_errmsg);
        xml_parser_errmsg = NULL;
    }

    doc = xmlParseMemory(contents, strlen(contents));
    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        if (errmsg) {
            *errmsg = g_strdup(xml_parser_errmsg);
        }

        g_free(xml_parser_errmsg);
        xml_parser_errmsg = NULL;

        xmlFreeDoc(doc);
        return NULL;
    }

    if (xmlStrcmp(cur->name, (const xmlChar *) "NNTPGrab")) {
        if (errmsg) {
            *errmsg = g_strdup("parse_version_info: document of the wrong type, root node != NNTPGrab");
        }

        xmlFreeDoc(doc);

        return FALSE;
    }

    cur = cur->xmlChildrenNode;
    if (!cur) {
        if (errmsg) {
            *errmsg = g_strdup("parse_version_info: no children of root-node!");
        }

        xmlFreeDoc(doc);

        return FALSE;
    }

    info = g_slice_new0(UpdateInfo);

    do {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "VersionInfo")) {
            parse_info(doc, cur->xmlChildrenNode, info);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "ChangeLog")) {
            parse_changelogs(doc, cur->xmlChildrenNode, info);
        }
    } while ((cur=cur->next));

    xmlFreeDoc(doc);

    return info;
}

#define HTML_INFO_HEADER                                            \
"<HTML><HEAD></HEAD><BODY>"                                         \
"<CENTER><H1>%s</H1></CENTER><BR>"

#define HTML_INFO_VERSION                                           \
"<CENTER>"                                                          \
"<TABLE WIDTH=\"50%%\">"                                            \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>"                             \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>"                             \
"</TABLE>"                                                          \
"</CENTER><BR>"

#define HTML_INFO_UNSTABLE_VERSION                                  \
"<CENTER>"                                                          \
"<TABLE WIDTH=\"50%%\">"                                            \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>"                             \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>"                             \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>"                             \
"</TABLE>"                                                          \
"</CENTER><BR>"

#define HTML_INFO_NO_UPDATE                                         \
"<CENTER>"                                                          \
"<P>%s</P>"                                                         \
"</CENTER>"

#define HTML_INFO_UNSTABLE_NO_UPDATE                                \
"<CENTER>"                                                          \
"<P>%s<BR/>%s</P>"                                                  \
"</CENTER>"

#define HTML_INFO_UPDATE                                            \
"<CENTER>"                                                          \
"<H3>%s</H3><BR>"                                                   \
"%s %i.%i.%i %s %s %s"                                              \
"</CENTER>"

#define HTML_INFO_CHANGELOG_BEGIN                                   \
"<P><U><B>%s %i.%i.%i:</B></U>"                                           \
"<UL>"

#define HTML_INFO_CHANGELOG_END                                     \
"</UL></P>"

#define HTML_UNSTABLE_NOTICE                                        \
"<CENTER><P><B>%s</B></P></CENTER>"

#define HTML_INFO_WEBSITE                                           \
"<CENTER><P>%s"                                                     \
"<A HREF=\"http://www.nntpgrab.nl\">http://www.nntpgrab.nl</A>"     \
"</P></CENTER>"

#define CHECK_VERSION_IS_OLDER(major,minor,micro)    \
    (MAJOR_VERSION < (major) || \
     (MAJOR_VERSION == (major) && MINOR_VERSION < (minor)) || \
     (MAJOR_VERSION == (major) && MINOR_VERSION == (minor) && \
      MICRO_VERSION < (micro)))

#define CHECK_IS_VERSION(major,minor,micro)    \
     (MAJOR_VERSION == (major) && MINOR_VERSION == (minor) && MICRO_VERSION == (micro))

void
generate_html(UpdateInfo *info)
{
    GtkWidget *html;
    GtkHTMLStream *stream;
    char *data;

    html = nntpgrab_gui_base_get_widget("htmlInformation");
    stream = gtk_html_begin(GTK_HTML(html));

    data = g_strdup_printf(HTML_INFO_HEADER, _("Welcome to NNTPGrab"));
    gtk_html_write(GTK_HTML(html), stream, data, -1);
    g_free(data);

    if (info->unstable_major_version >= 0) {
        data = g_strdup_printf(HTML_INFO_UNSTABLE_VERSION, _("You are using version:"), MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION, _("The latest available stable version is:"), info->stable_major_version, info->stable_minor_version, info->stable_micro_version, _("The latest available unstable version is:"), info->unstable_major_version, info->unstable_minor_version, info->unstable_micro_version);
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);
    } else {
        data = g_strdup_printf(HTML_INFO_VERSION, _("You are using version:"), MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION, _("The latest available version is:"), info->stable_major_version, info->stable_minor_version, info->stable_micro_version);
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);
    }

    if (info->unstable_major_version < 0 && CHECK_IS_VERSION(info->stable_major_version, info->stable_minor_version, info->stable_micro_version)) {
        data = g_strdup_printf(HTML_INFO_NO_UPDATE, _("You are using the latest version of NNTPGrab"));
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);

        data = g_strdup_printf(HTML_INFO_WEBSITE, _("For more information, go to our website "));
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);
    } else if (CHECK_IS_VERSION(info->unstable_major_version, info->unstable_minor_version, info->unstable_micro_version)) {
        data = g_strdup_printf(HTML_INFO_UNSTABLE_NO_UPDATE, _("You are using the latest unstable version of NNTPGrab"), _("If you find any bugs with this version, please report these"));
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);

        data = g_strdup_printf(HTML_INFO_WEBSITE, _("For more information, go to our website "));
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);
    } else if ((info->unstable_major_version < 0 && !(CHECK_VERSION_IS_OLDER(info->stable_major_version, info->stable_minor_version, info->stable_micro_version))) ||
               (info->unstable_major_version >=0 && !(CHECK_VERSION_IS_OLDER(info->unstable_major_version, info->unstable_minor_version, info->unstable_micro_version)))) {
        data = g_strdup_printf(HTML_INFO_NO_UPDATE, _("You are using a snapshot version of NNTPGrab"));
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);

        data = g_strdup_printf(HTML_INFO_WEBSITE, _("For more information, go to our website "));
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);
    } else {
        GList *list;
        GList *list2;

        if (info->unstable_major_version >= 0) {
            data = g_strdup_printf(HTML_INFO_UPDATE, _("A new unstable version of NNTPGrab is available!"), _("Version"), info->unstable_major_version, info->unstable_minor_version, info->unstable_micro_version, _("is released on"), info->date_latest_unstable_version_str, _("and contains the following changes :"));
        } else {
            data = g_strdup_printf(HTML_INFO_UPDATE, _("A new version of NNTPGrab is available!"), _("Version"), info->stable_major_version, info->stable_minor_version, info->stable_micro_version, _("is released on"), info->date_latest_stable_version_str, _("and contains the following changes :"));
        }
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);

        list = info->changelogs;
        while (list) {
            ChangeLog *changelog = list->data;

            if (!CHECK_VERSION_IS_OLDER(changelog->major_version, changelog->minor_version, changelog->micro_version)) {
                list = g_list_next(list);
                continue;
            }

            data = g_strdup_printf(HTML_INFO_CHANGELOG_BEGIN, _("Changes in version"), changelog->major_version, changelog->minor_version, changelog->micro_version);
            gtk_html_write(GTK_HTML(html), stream, data, -1);
            g_free(data);

            list2 = changelog->entries;
            while (list2) {
                char *entry = list2->data;

                data = g_strdup_printf("<LI>%s", entry);
                gtk_html_write(GTK_HTML(html), stream, data, -1);
                g_free(data);

                list2 = g_list_next(list2);
            }

            data = g_strdup_printf(HTML_INFO_CHANGELOG_END);
            gtk_html_write(GTK_HTML(html), stream, data, -1);
            g_free(data);

            list = g_list_next(list);
        }

        if (info->unstable_major_version >= 0) {
            data = g_strdup_printf(HTML_UNSTABLE_NOTICE, _("Please note that the unstable version may still contain bugs, so only use the unstable version when you want to help out testing"));
            gtk_html_write(GTK_HTML(html), stream, data, -1);
            g_free(data);
        }

        data = g_strdup_printf(HTML_INFO_WEBSITE, _("To retrieve the latest version and for more information, go to our website "));
        gtk_html_write(GTK_HTML(html), stream, data, -1);
        g_free(data);
    }

    gtk_html_end(GTK_HTML(html), stream, GTK_HTML_STREAM_OK);
}

static gboolean
process_update_info(gpointer data)
{
    UpdateInfo *info;
    GList *list;
    char *errmsg = NULL;
    char *contents = (char*) data;

    if (!contents) {
        // Update check failed
        GtkWidget *windowMain = nntpgrab_gui_base_get_widget("windowMain");
        nntpgrab_gui_base_dialog_show(windowMain, _("Update check failed"), GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);
        return FALSE;
    }

    info = parse_version_info(contents, &errmsg);

    g_free(contents);

    if (!info) {
        // Invalid data received
        char *msg;
        GtkWidget *windowMain = nntpgrab_gui_base_get_widget("windowMain");

        msg = g_strdup_printf(_("Update check failed due to invalid data:\n%s"), errmsg);
        nntpgrab_gui_base_dialog_show(windowMain, msg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);

        g_free(msg);
        g_free(errmsg);

        return FALSE;
    }

    generate_html(info);

    list = info->changelogs;
    while (list) {
        ChangeLog *changelog = list->data;
        GList *list2;

        list2 = changelog->entries;
        while (list2) {
            g_free(list2->data);
            list2 = g_list_next(list2);
        }
        g_list_free(changelog->entries);

        g_slice_free(ChangeLog, list->data);
        list = g_list_next(list);
    }
    g_list_free(info->changelogs);
    g_slice_free(UpdateInfo, info);

    return FALSE;
}

static gpointer
check_for_updates(gpointer data)
{
    char *contents;

    contents = retrieve_website_info();

    g_idle_add(process_update_info, contents);

    return NULL;
}

void
info_initialize()
{
    g_thread_create(check_for_updates, NULL, TRUE, NULL);
}

G_MODULE_EXPORT void
on_btnDonate_clicked(GtkWidget *caller, gpointer data)
{
    open_uri(NULL, "http://www.nntpgrab.nl/donate.php", NULL);
}
