/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <cairo.h>
#include "gui.h"
#include "nntpgrab_utils.h"

static int bytes_received[10];
static time_t stamp = 0;
static double average_speed = 0;

static gboolean
expose_event_callback (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
    cairo_t *cr = gdk_cairo_create(GDK_DRAWABLE(widget->window));
    cairo_text_extents_t te;
    int i;
    int max = 0;
    int width = widget->allocation.width;
    int height = widget->allocation.height - (widget->allocation.height / 10);  // Take 10% off the top
    double y_factor;
    int y_offset = widget->allocation.height / 10;  // 10%
    double x, y;
    char *msg;
    char *avg_speed;

    // Calculate the peak value
    for (i = 0; i < 10; i++) {
        if (bytes_received[i] > max) {
            max = bytes_received[i];
        }
    }

    // Draw the background
    cairo_set_line_width (cr, 1.0);
    cairo_set_source_rgb (cr, 255, 255, 255);
    cairo_rectangle (cr, 0, 0, width, widget->allocation.height);
    cairo_fill (cr);

    if (max == 0) {
        // Nothing to draw yet
        cairo_destroy(cr);
        return TRUE;
    }

    y_factor = height;
    y_factor /= max;

    cairo_set_source_rgb (cr, 0.565, 0.7, 0.87);

    // Create a new path in which all the coordinates are defined
    cairo_new_path(cr);

    // Use the first coordinate as starting point
    y = bytes_received[0];
    y *= y_factor;
    y = height - y;
    y += y_offset;

    cairo_move_to (cr, 0, y);

    // Create a path through the rest of the coordinates
    for (i = 1; i < 10; i++) {
        x = width;
        x /= 9;
        x *= i;

        y = bytes_received[i];
        y *= y_factor;
        y = height - y;
        y += y_offset;

        cairo_line_to (cr, x, y);
    }

    // Finalize the path by going to the bottom right corner
    cairo_line_to(cr, width, height + y_offset);

    // and to the bottom right corner
    cairo_line_to(cr, 0, height + y_offset);

    // Connect back to the first coordinate
    cairo_close_path(cr);

    // Fill the path
    cairo_set_source_rgb (cr, 0.565, 0.7, 0.87);
    cairo_fill (cr);

    // Draw the average speed
    avg_speed = nntpgrab_utils_calculate_file_size((guint64) average_speed);
    msg = g_strdup_printf("%s/sec", avg_speed);

    cairo_set_font_size(cr, 12.0);
    cairo_text_extents (cr, msg, &te);
    cairo_move_to (cr, width - te.width - te.x_bearing - 5, height);
    cairo_set_source_rgb (cr, 0, 0, 0);
    cairo_show_text (cr, msg);

    g_free(avg_speed);
    g_free(msg);

    cairo_destroy(cr);

    return TRUE;
}

static void
traffic_monitor_update_cb (NNTPGrabCore *core, int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, guint64 last_update_stamp, double average)
{
    static GtkWidget *windowMain = NULL;
    static GtkWidget *drawingareaSpeed = NULL;

    if (!windowMain) {
        windowMain = nntpgrab_gui_base_get_widget("windowMain");
    }

    if (!drawingareaSpeed) {
        drawingareaSpeed = nntpgrab_gui_base_get_widget("drawingareaSpeed");
    }

    bytes_received[0] = bytes_received1;
    bytes_received[1] = bytes_received2;
    bytes_received[2] = bytes_received3;
    bytes_received[3] = bytes_received4;
    bytes_received[4] = bytes_received5;
    bytes_received[5] = bytes_received6;
    bytes_received[6] = bytes_received7;
    bytes_received[7] = bytes_received8;
    bytes_received[8] = bytes_received9;
    bytes_received[9] = bytes_received10;

    stamp = last_update_stamp;
    average_speed = average;

    if (gdk_window_get_state(windowMain->window) != GDK_WINDOW_STATE_ICONIFIED) {
        gdk_window_invalidate_rect(drawingareaSpeed->window, NULL, FALSE);
    }
}

void
monitor_initialize(void)
{
    GtkWidget *drawingareaSpeed = nntpgrab_gui_base_get_widget("drawingareaSpeed");

    g_signal_connect (G_OBJECT (drawingareaSpeed), "expose_event", G_CALLBACK (expose_event_callback), NULL);

    g_signal_connect(core, "traffic_monitor_update", G_CALLBACK(traffic_monitor_update_cb), NULL);
}
