/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef WIN32
#include <windows.h>
#include <gdk/gdkwin32.h>
#endif
#include <stdlib.h>
#include <string.h>
#include "gui.h"
#include <gdk/gdkkeysyms.h>
#include "config_gui.h"
#include "nntpgrab_utils.h"

#ifdef HAVE_LIBNOTIFY
#include <libnotify/notify.h>
#endif

static gboolean find_iter_by_collection_name(GtkTreeModel *model, const char *collection_name, GtkTreeIter *iter);
static gboolean find_iter_by_subject(GtkTreeModel *model, GtkTreeIter parent, const char *subject, GtkTreeIter *iter);

static SchedularState status_flag = SCHEDULAR_STATE_RUNNING;     // Schedular is running by default

enum {
    FIELD_COLLECTION_NAME,
    FIELD_SUBJECT,
    FIELD_POSTER,
    FIELD_GROUPS,
    FIELD_STAMP_STR,
    FIELD_PROGRESS,
    FIELD_FILESIZE,
    FIELD_FILESIZE_STR,
    FIELD_FILESIZE_REMAINING,
    FIELD_FILESIZE_REMAINING_STR,
    FIELD_NUM_PARTS,
    FIELD_PARTS_DONE,
    FIELD_PARTS_FAILED,
    FIELD_STATE_STR,
    FIELD_ESTIMATED_TIME_REMAINING,
    FIELD_ESTIMATED_TIME_TO_FINISH,
    FIELD_POSITION,
    FIELD_REAL_FILENAME,
    LAST_FIELD
};
static GtkTreeViewColumn *columns[LAST_FIELD] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 , 0 };

#ifdef WIN32
static void
open_file(const char *file, gboolean open_folder)
{
    char *mode;
    char *file_localized;
    GtkWidget *windowMain = nntpgrab_gui_base_get_widget("windowMain");

    if (open_folder) {
        mode = "explore";
    } else {
        mode = "open";
    }

    file_localized = g_win32_locale_filename_from_utf8(file);
    g_return_if_fail(file_localized != NULL);

    if ((int) ShellExecuteA((HWND) GDK_WINDOW_HWND(windowMain->window), mode, file_localized, NULL, file_localized, SW_SHOW) < 32) {
        char *msg;
        GtkWidget *windowMain;

        windowMain = nntpgrab_gui_base_get_widget("windowMain");
        if (open_folder) {
            msg = g_strdup_printf(_("Unable to navigate to the download folder\n"));    // TODO: Add GetLastError message
        } else {
            msg = g_strdup_printf(_("Unable to execute file\n%s"), file);    // TODO: Add GetLastError message
        }
        nntpgrab_gui_base_dialog_show(windowMain, msg, GTK_MESSAGE_INFO, GTK_BUTTONS_OK);
        g_free(msg);
    }

    g_free(file_localized);
}
#else
static void
open_file(const char *folder_unquoted, gboolean open_folder)
{
    char *cmd = NULL;
    GError *err = NULL;
    char *folder = g_shell_quote(folder_unquoted);
    gboolean ret;

#if GTK_CHECK_VERSION(2,14,0) && !defined(DARWIN)
    char *uri = g_strdup_printf("file://%s", folder_unquoted);
    ret = gtk_show_uri(NULL, uri, GDK_CURRENT_TIME, &err);
    g_free(uri);
#else
#ifdef DARWIN
    cmd = g_strdup_printf("open %s", folder);
#else
    cmd = g_strdup_printf("xdg-open %s", folder);
#endif

    ret = g_spawn_command_line_async(cmd, &err);
#endif /* GTK_CHECK_VERSION(2,14,0) */

    if (!ret) {
        char *msg;
        GtkWidget *windowMain;

        windowMain = nntpgrab_gui_base_get_widget("windowMain");
        if (open_folder) {
            msg = g_strdup_printf(_("Unable to navigate to the download folder:\n%s"), err->message);
        } else {
            msg = g_strdup_printf(_("Unable to execute file\n%s:\n%s"), folder, err->message);
        }
        g_error_free(err);
        nntpgrab_gui_base_dialog_show(windowMain, msg, GTK_MESSAGE_INFO, GTK_BUTTONS_OK);
        g_free(msg);
    }

    g_free(cmd);
}
#endif

G_MODULE_EXPORT gboolean
on_treeDownloadQueue_button_press_event(GtkWidget *caller, GdkEventButton *event, gpointer data)
{
    GtkWidget *menuDownloadQueue;
    GtkTreeModel *model;
    GtkTreeSelection *selection;
#if GTK_CHECK_VERSION(2,12,0)
    GObject *menuitemExecuteItem;
    GValue val;
#else
    GtkWidget *menuitemExecuteItem;
#endif
#ifdef NNTPGRAB_NETWORKED
#if GTK_CHECK_VERSION(2,12,0)
    GObject *menuitemOpenDownloadFolder;
#else
    GtkWidget *menuitemOpenDownloadFolder;
#endif
#endif

    // We only want to catch the right mouse button
    if (event->button != 3) {
        return FALSE;
    }

    // Is at least one item selected?
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(caller));
    if (gtk_tree_selection_count_selected_rows(selection) == 0)
    {
        GtkTreePath *path;

        /* Get tree path for row that was clicked */
        if (!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(caller),
                                            (gint) event->x,
                                            (gint) event->y,
                                            &path, NULL, NULL, NULL))
        {
            // No item(s) selected and no item under the cursor. Ignore
            return FALSE;
        }

        gtk_tree_selection_unselect_all(selection);
        gtk_tree_selection_select_path(selection, path);
        gtk_tree_path_free(path);
    }

    // Do we have exactly one selected item and is this item already downloaded?
#if GTK_CHECK_VERSION(2,12,0)
    menuitemExecuteItem = nntpgrab_gui_base_get_object("menuitemExecuteItem");
    memset(&val, 0, sizeof(val));
    g_value_init(&val, G_TYPE_BOOLEAN);
    g_value_set_boolean (&val, FALSE);
    g_object_set_property(G_OBJECT(menuitemExecuteItem), "sensitive", &val);
#else
    menuitemExecuteItem = nntpgrab_gui_base_get_widget("menuitemExecuteItem");
    gtk_widget_set_sensitive(menuitemExecuteItem, FALSE);
#endif

#ifdef NNTPGRAB_NETWORKED
#if GTK_CHECK_VERSION(2,12,0)
    menuitemOpenDownloadFolder = nntpgrab_gui_base_get_object("menuitemOpenDownloadFolder");
    memset(&val, 0, sizeof(val));
    g_value_init(&val, G_TYPE_BOOLEAN);
    g_value_set_boolean (&val, FALSE);
    g_object_set_property(G_OBJECT(menuitemOpenDownloadFolder), "sensitive", &val);
#else
    menuitemOpenDownloadFolder = nntpgrab_gui_base_get_widget("menuitemOpenDownloadFolder");
    gtk_widget_set_sensitive(menuitemOpenDownloadFolder, FALSE);
#endif
#endif

    if (gtk_tree_selection_count_selected_rows(selection) == 1) {
        char *progress_str = NULL;
        GtkTreeIter iter;
        GList *list = gtk_tree_selection_get_selected_rows(selection, &model);

        gtk_tree_model_get_iter(model, &iter, (GtkTreePath *) list->data);
        gtk_tree_model_get(model, &iter, FIELD_STATE_STR, &progress_str, -1);

        if (progress_str != NULL) {
            if (!strcmp(progress_str, _("Done")) ||
                !strcmp(progress_str, _("Incomplete"))) {

#ifndef NNTPGRAB_NETWORKED
#if GTK_CHECK_VERSION(2,12,0)
                g_value_set_boolean (&val, TRUE);
                g_object_set_property(G_OBJECT(menuitemExecuteItem), "sensitive", &val);
#else
                gtk_widget_set_sensitive(menuitemExecuteItem, TRUE);
#endif
#endif
            }

            g_free(progress_str);
        }

        g_list_foreach (list, (GFunc) gtk_tree_path_free, NULL);
        g_list_free (list);
    }

    // Do we have multiple items selected and do these all belong to the same collection ?
    // TODO

    menuDownloadQueue = nntpgrab_gui_base_get_widget("menuDownloadQueue");
    gtk_menu_popup(GTK_MENU(menuDownloadQueue), NULL, NULL, NULL, NULL, event->button, event->time);

    return TRUE;
}

G_MODULE_EXPORT void
on_menuitemRestartTask_activate(GtkWidget *caller, gpointer data)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *list, *selectedRows;
    GtkTreeModel *model;
    GtkTreeIter iter, parent;
    char *collection_name = NULL;
    char *subject = NULL;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_assert(gtk_tree_selection_count_selected_rows(selection) > 0);

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    // Duplicate the list and transform the items to GtkTreeRowReferences
    // This is because we are going to delete items in the tree
    list = selectedRows;
    while (list) {
        GtkTreePath *path = list->data;

        gtk_tree_model_get_iter(model, &iter, path);
        if (!gtk_tree_model_iter_parent(model, &parent, &iter)) {
            // Item contains no parent, restart the whole collection
            gtk_tree_model_get(model, &iter, FIELD_COLLECTION_NAME, &collection_name, -1);
            nntpgrab_schedular_restart_task(core, collection_name, NULL, NULL);

            g_free(collection_name);
        } else {
            // Restart one item from the collection
            gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &collection_name, -1);
            gtk_tree_model_get(model, &iter, FIELD_SUBJECT, &subject, -1);

            nntpgrab_schedular_restart_task(core, collection_name, subject, NULL);

            g_free(collection_name);
            g_free(subject);
        }

        list = g_list_next(list);
    }

    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);

    nntpgrab_schedular_save_queue(core, NULL);
}

G_MODULE_EXPORT void
on_menuitemOpenDownloadFolder_activate(GtkWidget *caller, gpointer data)
{
    // Is only one collection selected or more ?
    gboolean multiple_collections = FALSE;
    char *collection_name = NULL;
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *list, *selectedRows;
    GtkTreeModel *model;
    ConfigOpts opts;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_assert(gtk_tree_selection_count_selected_rows(selection) > 0);

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    list = selectedRows;
    while (list) {
        GtkTreePath *path = list->data;
        GtkTreeIter iter;
        char *name = NULL;

        gtk_tree_model_get_iter(model, &iter, path);
        gtk_tree_model_get(model, &iter, FIELD_COLLECTION_NAME, &name, -1);

        if (name == NULL || strlen(name) == 0) {
            GtkTreeIter parent;

            if (name) {
                g_free(name);
            }

            // Look at the parent
            gtk_tree_model_iter_parent(model, &parent, &iter);
            gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &name, -1);
        }

        g_assert(name);

        if (!collection_name) {
            collection_name = g_strdup(name);
        } else {
            if (strcmp(name, collection_name)) {
                // A previous found collection name and this collection name differ!
                multiple_collections = TRUE;
            }
        }

        g_free(name);

        list = g_list_next(list);
    }

    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);

    opts = nntpgrab_config_get_opts(core);

    if (multiple_collections) {
        // Open the default download folder
        open_file(opts.download_directory, TRUE);
    } else {
        // Open the download folder of the active collection
        char *folder = g_strdup_printf("%s%s%s", opts.download_directory, G_DIR_SEPARATOR_S, collection_name);

        if (!g_file_test(folder, G_FILE_TEST_IS_DIR)) {
            // Folder doesn't exist yet, fall back to default download directory
            open_file(opts.download_directory, TRUE);
        } else {
            open_file(folder, TRUE);
        }

        g_free(folder);
    }

    if (collection_name) {
        g_free(collection_name);
    }
}

G_MODULE_EXPORT void
on_menuitemExecuteItem_activate(GtkWidget *caller, gpointer data)
{
    char *collection_name = NULL;
    char *real_filename = NULL;
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *selectedRows;
    GtkTreeModel *model;
    ConfigOpts opts;
    GtkTreePath *path;
    GtkTreeIter iter;
    char *subject;
    char *filename = NULL;
    char *full_filename;
    GtkTreeIter parent;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(gtk_tree_selection_count_selected_rows(selection) == 1);

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    path = selectedRows->data;

    gtk_tree_model_get_iter(model, &iter, path);
    gtk_tree_model_get(model, &iter, FIELD_SUBJECT, &subject, -1);

    // Look at the parent
    gtk_tree_model_iter_parent(model, &parent, &iter);
    gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &collection_name, -1);
    gtk_tree_model_get(model, &iter, FIELD_REAL_FILENAME, &real_filename, -1);

    g_return_if_fail(real_filename != NULL);

    opts = nntpgrab_config_get_opts(core);
    full_filename = g_strdup_printf("%s%s%s%s%s", opts.download_directory, G_DIR_SEPARATOR_S, collection_name, G_DIR_SEPARATOR_S, real_filename);

    open_file(full_filename, FALSE);

    g_free(full_filename);
    g_free(subject);
    g_free(real_filename);
    g_free(collection_name);
    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);
}

typedef enum {
    MOVE_DIRECTION_TOP,
    MOVE_DIRECTION_UP,
    MOVE_DIRECTION_DOWN,
    MOVE_DIRECTION_BOTTOM
} move_direction;

static void
task_moved_cb(NNTPGrabCore *core, const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position, gpointer data)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter parent, orig_iter, new_iter;
    int i;

#if 0
    g_print("%s: orig_collection_name = %s\n", __PRETTY_FUNCTION__, orig_collection_name);
    g_print("%s: subject = %s\n", __PRETTY_FUNCTION__, subject);
    g_print("%s: new_collection_name = %s\n", __PRETTY_FUNCTION__, new_collection_name);
    g_print("%s: old_position = %i\n", __PRETTY_FUNCTION__, old_position);
    g_print("%s: new_position = %i\n", __PRETTY_FUNCTION__, new_position);
#endif

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), new_collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &orig_iter));

    if (new_position == -1) {
        new_position = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(store), &parent);

        i = 0;
        gtk_tree_model_iter_children(GTK_TREE_MODEL(store), &new_iter, &parent);
        while (i != new_position) {
            gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &new_iter);
            i++;
        }

        gtk_tree_store_move_after(GTK_TREE_STORE(store), &orig_iter, &new_iter);
    } else {
        i = 0;
        gtk_tree_model_iter_children(GTK_TREE_MODEL(store), &new_iter, &parent);
        while (i != new_position) {
            gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &new_iter);
            i++;
        }

        if (old_position < new_position) {
            gtk_tree_store_move_after(GTK_TREE_STORE(store), &orig_iter, &new_iter);
        } else {
            gtk_tree_store_move_before(GTK_TREE_STORE(store), &orig_iter, &new_iter);
        }
    }
}

static void
collection_moved_cb(NNTPGrabCore *core, const char *collection_name, int old_position, int new_position)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter orig_iter, new_iter;
    int i;

#if 0
    g_print("%s: collection_name = %s\n", __PRETTY_FUNCTION__, collection_name);
    g_print("%s: old_position = %i\n", __PRETTY_FUNCTION__, old_position);
    g_print("%s: new_position = %i\n", __PRETTY_FUNCTION__, new_position);
#endif

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &orig_iter));

    if (new_position == -1) {
        new_position = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(store), NULL);

        i = 0;
        gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &new_iter);
        while (i != new_position) {
            gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &new_iter);
            i++;
        }

        gtk_tree_store_move_after(GTK_TREE_STORE(store), &orig_iter, &new_iter);
    } else {
        i = 0;
        gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &new_iter);
        while (i != new_position) {
            gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &new_iter);
            i++;
        }

        if (old_position < new_position) {
            gtk_tree_store_move_after(GTK_TREE_STORE(store), &orig_iter, &new_iter);
        } else {
            gtk_tree_store_move_before(GTK_TREE_STORE(store), &orig_iter, &new_iter);
        }
    }
}

static void
move_item(move_direction direction)
{
    char *collection_name = NULL;
    char *subject = NULL;
    int position = -1;
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *list, *selectedRows;
    GtkTreeModel *model;
    GList *references = NULL;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(gtk_tree_selection_count_selected_rows(selection) > 0);

    // Duplicate the list with selected rows into GtkTreeRowReference's as the position of
    // selected rows can change during the loop later on

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    list = selectedRows;
    while (list) {
        GtkTreePath *path = list->data;
        GtkTreeRowReference *ref;

        ref = gtk_tree_row_reference_new(model, path);
        references = g_list_append(references, ref);

        list = g_list_next(list);
    }

    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);

    if (direction == MOVE_DIRECTION_DOWN || direction == MOVE_DIRECTION_TOP) {
        references = g_list_reverse(references);
    }

    list = references;
    while (list) {
        GtkTreeRowReference *ref = list->data;
        GtkTreePath *path;
        GtkTreeIter iter;
        gboolean ret = FALSE;
        char *path_str;
        char **values;
        gboolean is_child = FALSE;

        if (collection_name) {
            g_free(collection_name);
            collection_name = NULL;
        }

        if (subject) {
            g_free(subject);
            subject = NULL;
        }

        path = gtk_tree_row_reference_get_path(ref);
        if (!path) {
            list = g_list_next(list);
            gtk_tree_row_reference_free(ref);
            continue;
        }

        gtk_tree_model_get_iter(model, &iter, path);
        gtk_tree_model_get(model, &iter, FIELD_COLLECTION_NAME, &collection_name, FIELD_SUBJECT, &subject, -1);

        path_str = gtk_tree_path_to_string(path);
        values = g_strsplit(path_str, ":", 0);
        g_return_if_fail(values != NULL);
        if (values[0]) {
            position = atoi(values[0]);
            if (values[1]) {
                position = atoi(values[1]);
            }
        }

        g_strfreev(values);
        g_free(path_str);

        if (collection_name == NULL || strlen(collection_name) == 0) {
            GtkTreeIter parent;

            // Look at the parent
            gtk_tree_model_iter_parent(model, &parent, &iter);
            gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &collection_name, -1);

            is_child = TRUE;
        }

        g_assert(collection_name);

        switch (direction) {
            case MOVE_DIRECTION_TOP:
                if (is_child) {
                    ret = nntpgrab_schedular_move_task(core, collection_name, subject, collection_name, 0);
                } else {
                    ret = nntpgrab_schedular_move_collection(core, collection_name, 0);
                }
                break;

            case MOVE_DIRECTION_UP:
                if (position == 0) {
                    break;
                }
                if (is_child) {
                    ret = nntpgrab_schedular_move_task(core, collection_name, subject, collection_name, position - 1);
                } else {
                    ret = nntpgrab_schedular_move_collection(core, collection_name, position - 1);
                }
                break;

            case MOVE_DIRECTION_DOWN:
                if (is_child) {
                    ret = nntpgrab_schedular_move_task(core, collection_name, subject, collection_name, position + 1);
                } else {
                    ret = nntpgrab_schedular_move_collection(core, collection_name, position + 1);
                }
                break;

            case MOVE_DIRECTION_BOTTOM:
                if (is_child) {
                    ret = nntpgrab_schedular_move_task(core, collection_name, subject, collection_name, -1);
                } else {
                    ret = nntpgrab_schedular_move_collection(core, collection_name, -1);
                }
                break;

            default:
                break;
        }

        list = g_list_next(list);
        gtk_tree_path_free(path);
        gtk_tree_row_reference_free(ref);
    }

    g_list_free(references);

    nntpgrab_schedular_save_queue(core, NULL);

    if (collection_name) {
        g_free(collection_name);
    }

    if (subject) {
        g_free(subject);
    }
}

G_MODULE_EXPORT void
on_menuitemMoveToTop_activate(GtkWidget *caller, gpointer data)
{
    move_item(MOVE_DIRECTION_TOP);
}

G_MODULE_EXPORT void
on_menuitemMoveUp_activate(GtkWidget *caller, gpointer data)
{
    move_item(MOVE_DIRECTION_UP);
}

G_MODULE_EXPORT void
on_menuitemMoveDown_activate(GtkWidget *caller, gpointer data)
{
    move_item(MOVE_DIRECTION_DOWN);
}

G_MODULE_EXPORT void
on_menuitemMoveToBottom_activate(GtkWidget *caller, gpointer data)
{
    move_item(MOVE_DIRECTION_BOTTOM);
}

G_MODULE_EXPORT void
on_btnRemoveItemFromQueue_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *list, *selectedRows, *list_new;
    GtkTreeModel *model;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(gtk_tree_selection_count_selected_rows(selection) > 0);

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    // Duplicate the list and transform the items to GtkTreeRowReferences
    // This is because we are going to delete items in the tree
    list = selectedRows;
    list_new = NULL;
    while (list) {
        GtkTreePath *path = list->data;

        list_new = g_list_append(list_new, gtk_tree_row_reference_new(model, path));

        list = g_list_next(list);
    }

    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);

    // Now loop trough all the items and remove the downloads tasks
    list = list_new;
    while (list) {
        GtkTreeRowReference *ref = list->data;

        if (gtk_tree_row_reference_valid(ref)) {
            GtkTreePath *path = gtk_tree_row_reference_get_path(ref);
            GtkTreeIter iter, parent;
            char *collection_name = NULL;
            char *subject = NULL;

            gtk_tree_model_get_iter(model, &iter, path);

            if (!gtk_tree_model_iter_parent(model, &parent, &iter)) {
                // Item contains no parent, delete the whole collection
                gtk_tree_model_get(model, &iter, FIELD_COLLECTION_NAME, &collection_name, -1);
                nntpgrab_schedular_del_task_from_queue(core, collection_name, NULL, NULL);

                g_free(collection_name);
            } else {
                // Delete one item from the collection
                gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &collection_name, -1);
                gtk_tree_model_get(model, &iter, FIELD_SUBJECT, &subject, -1);

                nntpgrab_schedular_del_task_from_queue(core, collection_name, subject, NULL);

                g_free(collection_name);
                g_free(subject);
            }
        }

        gtk_tree_row_reference_free(ref);

        list = g_list_next(list);
    }

    g_list_free(list_new);

    nntpgrab_schedular_save_queue(core, NULL);
}

static char *
generate_stamp(time_t stamp)
{
    char buf[64];
    struct tm *now;

    if ((now = localtime(&stamp)) == NULL) {
        // Date could not be parted
        return g_strdup("");
    }

    strftime(buf, sizeof(buf), "%c", now);

    return g_strdup(buf);
}

static int
calculate_progress(guint64 file_size, guint64 file_size_remaining)
{
    double progress;
    int progress_int;

    // Prevent a division by zero
    if (file_size_remaining == 0) {
        return 100;
    }

    progress = (double) file_size - file_size_remaining;
    progress /= file_size;
    progress *= 100;
    progress_int = (int) progress;

    if (progress_int > 100) {
        g_print(__FILE__ ":%i progress overflow!\n", __LINE__);
        g_print("file_size = %lli\n", file_size);
        g_print("file_size_remaining = %lli\n", file_size_remaining);
        progress_int = 100;
    }

    if (progress < 0) {
        g_print(__FILE__ ":%i progress underflow!\n", __LINE__);
        g_print("file_size = %lli\n", file_size);
        g_print("file_size_remaining = %lli\n", file_size_remaining);
        progress_int = 0;
    }

    return progress_int;
}

static void
file_download_state_update (NNTPGrabCore *core, const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_file_size, guint64 total_file_size_remaining, gpointer data)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;
    int progress;
    char *size_remaining;
    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    // Update the parent
    progress = calculate_progress(total_file_size, total_file_size_remaining);
    size_remaining = nntpgrab_utils_calculate_file_size(total_file_size_remaining);

    gtk_tree_store_set(GTK_TREE_STORE(store), &parent,  FIELD_PROGRESS, progress,
                                                        FIELD_FILESIZE_REMAINING, total_file_size_remaining,
                                                        FIELD_FILESIZE_REMAINING_STR, size_remaining, -1);

    g_free(size_remaining);

    // Update the child
    progress = calculate_progress(file_size, file_size_remaining);
    size_remaining = nntpgrab_utils_calculate_file_size(file_size_remaining);

    gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_PROGRESS, progress,
                                                        FIELD_PARTS_DONE, num_parts_done,
                                                        FIELD_PARTS_FAILED, num_parts_failed,
                                                        FIELD_FILESIZE_REMAINING, file_size_remaining,
                                                        FIELD_FILESIZE_REMAINING_STR, size_remaining, -1);

    g_free(size_remaining);
}

static void
collection_added (NNTPGrabCore *core, const char *collection_name)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    gtk_tree_store_append(GTK_TREE_STORE(store), &iter, NULL);
    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_COLLECTION_NAME, collection_name, FIELD_SUBJECT, collection_name, FIELD_STATE_STR, NULL, -1);
}

static void
collection_removed (NNTPGrabCore *core, const char *collection_name)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(store, collection_name, &iter));

    gtk_tree_store_remove(GTK_TREE_STORE(store), &iter);
}

static void
file_added (NNTPGrabCore *core, const char *collection_name, const char *subject, const char *poster, guint64 stamp, guint64 file_size, guint64 total_file_size, guint64 total_file_size_remaining, TaskState state, int num_parts, GList *groups)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter parent, iter;
    char *size;
    char *size_remaining;
    char *total_groups = NULL;
    char *stamp_str;
    GList *list;
    int progress;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(file_size > 0);

    list = groups;
    while (list) {
        if (total_groups) {
            char *tmp = g_strdup_printf("%s, %s", total_groups, (const char *) list->data);
            g_free(total_groups);
            total_groups = tmp;
        } else {
            total_groups = g_strdup(list->data);
        }

        list = g_list_next(list);
    }

    size = nntpgrab_utils_calculate_file_size(file_size);
    stamp_str = generate_stamp(stamp);
    gtk_tree_store_append(GTK_TREE_STORE(store), &iter, &parent);
    gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_SUBJECT, subject,
                                                        FIELD_POSTER, poster,
                                                        FIELD_GROUPS, total_groups,
                                                        FIELD_STAMP_STR, stamp_str,
                                                        FIELD_FILESIZE, file_size,
                                                        FIELD_FILESIZE_STR, size,
                                                        FIELD_FILESIZE_REMAINING, file_size,
                                                        FIELD_FILESIZE_REMAINING_STR, size,
                                                        FIELD_NUM_PARTS, num_parts,
                                                        FIELD_STATE_STR, NULL, -1);

    g_free(stamp_str);
    g_free(size);
    g_free(total_groups);

    // Update the parent
    size = nntpgrab_utils_calculate_file_size(total_file_size);
    size_remaining = nntpgrab_utils_calculate_file_size(total_file_size_remaining);
    progress = calculate_progress(total_file_size, total_file_size_remaining);

    gtk_tree_store_set(GTK_TREE_STORE(store), &parent,  FIELD_FILESIZE_STR, size,
                                                        FIELD_FILESIZE, total_file_size,
                                                        FIELD_FILESIZE_REMAINING_STR, size_remaining,
                                                        FIELD_FILESIZE_REMAINING, total_file_size_remaining,
                                                        FIELD_PROGRESS, progress, -1);

    g_free(size);
    g_free(size_remaining);
}

static void
file_removed (NNTPGrabCore *core, const char *collection_name, const char *subject, guint64 total_file_size, guint64 total_file_size_remaining, gpointer data)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;
    int progress;
    char *size, *size_remaining;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    size = nntpgrab_utils_calculate_file_size(total_file_size);
    size_remaining = nntpgrab_utils_calculate_file_size(total_file_size_remaining);
    progress = calculate_progress(total_file_size, total_file_size_remaining);

    gtk_tree_store_set(GTK_TREE_STORE(store), &parent,  FIELD_FILESIZE_STR, size,
                                                        FIELD_FILESIZE, total_file_size,
                                                        FIELD_FILESIZE_REMAINING_STR, size_remaining,
                                                        FIELD_FILESIZE_REMAINING, total_file_size_remaining,
                                                        FIELD_PROGRESS, progress, -1);

    gtk_tree_store_remove(GTK_TREE_STORE(store), &iter);

    g_free(size_remaining);
    g_free(size);
}

static void
file_state_changed (NNTPGrabCore *core, const char *collection_name, const char *subject, const char *real_filename, TaskState old_state, TaskState new_state)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;
    int progress;
    guint64 file_size;
    guint64 file_size_remaining;
    guint64 total_size;
    guint64 total_size_remaining;
    char *size_remaining;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    switch(new_state) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:       // The file was restarted
            // Update the row itself
            gtk_tree_model_get(GTK_TREE_MODEL(store), &iter,    FIELD_FILESIZE, &file_size,
                                                                FIELD_FILESIZE_REMAINING, &file_size_remaining, -1);

            size_remaining = nntpgrab_utils_calculate_file_size(file_size);

            gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_PARTS_DONE, (int) 0,
                                                                FIELD_PARTS_FAILED, (int) 0,
                                                                FIELD_PROGRESS, (int) 0,
                                                                FIELD_STATE_STR, NULL,
                                                                FIELD_FILESIZE_REMAINING, file_size,
                                                                FIELD_FILESIZE_REMAINING_STR, size_remaining, -1);

            g_free(size_remaining);

            // Update the parent
            gtk_tree_model_get(GTK_TREE_MODEL(store), &parent,  FIELD_FILESIZE, &total_size,
                                                                FIELD_FILESIZE_REMAINING, &total_size_remaining, -1);

            total_size_remaining += file_size - file_size_remaining;
            progress = calculate_progress(total_size, total_size_remaining);
            size_remaining = nntpgrab_utils_calculate_file_size(total_size_remaining);

            gtk_tree_store_set(GTK_TREE_STORE(store), &parent,  FIELD_PROGRESS, progress,
                                                                FIELD_FILESIZE_REMAINING, total_size_remaining,
                                                                FIELD_FILESIZE_REMAINING_STR, size_remaining, -1);

            g_free(size_remaining);

            break;

        case TASK_STATE_DOWNLOADING:
            gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_PROGRESS, (int) 0,
                                                                FIELD_STATE_STR, NULL, -1);     // The state message should be NULL. In that case the value of the progress bar will be shown
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
            size_remaining = nntpgrab_utils_calculate_file_size((guint64) 0);
            gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_PROGRESS, (int) 100,
                                                                FIELD_FILESIZE_REMAINING_STR, size_remaining,
                                                                FIELD_STATE_STR, _("Waiting"), -1);

            g_free(size_remaining);

            break;

        case TASK_STATE_DECODING:
            gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_PROGRESS, (int) 100,
                                                                FIELD_STATE_STR, _("Decoding"), -1);
            break;

        case TASK_STATE_FINISHED_COMPLETE:
            // Do we need to update the parent
            g_return_if_fail(real_filename != NULL);
            gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_STATE_STR, _("Done"),
                                                                FIELD_REAL_FILENAME, real_filename, -1);
            break;

        case TASK_STATE_FINISHED_INCOMPLETE:
            g_return_if_fail(real_filename != NULL);
            gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_STATE_STR, _("Incomplete"),
                                                                FIELD_REAL_FILENAME, real_filename, -1);
            break;

        case TASK_STATE_FINISHED_NO_PARTS_AVAIL:
            gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_STATE_STR, _("Not available"), -1);
            break;

        case TASK_STATE_SKIPPED:
            gtk_tree_store_set(GTK_TREE_STORE(store), &iter,    FIELD_STATE_STR, _("Skipped"),
                                                                FIELD_FILESIZE_REMAINING, (guint64) 0,
                                                                FIELD_FILESIZE_REMAINING_STR, NULL, -1);
            break;
    };
}

static void
foreach_collection_func (const char *collection_name, guint64 total_size, guint64 total_size_remaining, int position, gpointer data)
{
    GtkTreeStore *store = GTK_TREE_STORE(data);
    GtkTreeIter iter;
    char *size;
    char *size_remaining;
    int progress;

    size = nntpgrab_utils_calculate_file_size(total_size);
    size_remaining = nntpgrab_utils_calculate_file_size(total_size_remaining);
    progress = calculate_progress(total_size, total_size_remaining);

    gtk_tree_store_append(store, &iter, NULL);
    gtk_tree_store_set(store, &iter,    FIELD_COLLECTION_NAME, collection_name,
                                        FIELD_SUBJECT, collection_name,
                                        FIELD_FILESIZE_STR, size,
                                        FIELD_FILESIZE, total_size,
                                        FIELD_FILESIZE_REMAINING_STR, size_remaining,
                                        FIELD_FILESIZE_REMAINING, total_size_remaining,
                                        FIELD_STATE_STR, NULL,
                                        FIELD_PROGRESS, progress,
                                        FIELD_POSITION, position, -1);

    g_free(size_remaining);
    g_free(size);
}

static gboolean
find_iter_by_collection_name(GtkTreeModel *model, const char *collection_name, GtkTreeIter *iter)
{
    char *name;

    if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(model), iter)) {
        do {
            gtk_tree_model_get(GTK_TREE_MODEL(model), iter, FIELD_COLLECTION_NAME, &name, -1);

            if (!strcmp(collection_name, name)) {
                g_free(name);
                return TRUE;
            }

            g_free(name);
        } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(model), iter));
    }

    return FALSE;
}

static gboolean
find_iter_by_subject(GtkTreeModel *model, GtkTreeIter parent, const char *subject, GtkTreeIter *iter)
{
    char *name;

    if (!gtk_tree_model_iter_children(GTK_TREE_MODEL(model), iter, &parent)) {
        // Node contains no children. Shouldn't happen
        return FALSE;
    }

    do {
        gtk_tree_model_get(GTK_TREE_MODEL(model), iter, FIELD_SUBJECT, &name, -1);

        if (!strcmp(subject, name)) {
            g_free(name);
            return TRUE;
        }

        g_free(name);
    } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(model), iter));

    return FALSE;
}

static void
foreach_file_func (const char *collection_name, const char *subject, const char *poster, gint64 stamp, guint64 file_size, guint64 file_size_remaining, int position, int num_parts, int num_parts_downloaded, int num_parts_failed, TaskState status, const char *filename, gpointer data)
{
    GtkTreeStore *store = GTK_TREE_STORE(data);
    GtkTreeIter iter, parent;
    char *size;
    char *size_remaining;
    char *stamp_str;
    int progress;
    char *state_str;

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(file_size_remaining >= 0);

    size = nntpgrab_utils_calculate_file_size(file_size);
    size_remaining = nntpgrab_utils_calculate_file_size(file_size_remaining);
    stamp_str = generate_stamp(stamp);

    progress = calculate_progress(file_size, file_size_remaining);
    switch(status) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:
        case TASK_STATE_DOWNLOADING:
            state_str = NULL;
            break;

        case TASK_STATE_WAITING_FOR_DECODE:
            state_str = _("Waiting");
            break;

        case TASK_STATE_DECODING:
            state_str = _("Decoding");
            break;

        case TASK_STATE_FINISHED_COMPLETE:
            state_str = _("Done");
            break;

        case TASK_STATE_FINISHED_INCOMPLETE:
            state_str = _("Incomplete");
            break;

        case TASK_STATE_FINISHED_NO_PARTS_AVAIL:
            state_str = _("Not available");
            break;

        case TASK_STATE_SKIPPED:
            state_str = _("Skipped");
            break;

        default:
            state_str = NULL;
            break;
    };

    gtk_tree_store_append(store, &iter, &parent);
    gtk_tree_store_set(store, &iter,    FIELD_SUBJECT, subject,
                                        FIELD_POSTER, poster,
                                        FIELD_STAMP_STR, stamp_str,
                                        FIELD_FILESIZE, file_size,
                                        FIELD_FILESIZE_STR, size,
                                        FIELD_FILESIZE_REMAINING, file_size_remaining,
                                        FIELD_FILESIZE_REMAINING_STR, size_remaining,
                                        FIELD_NUM_PARTS, num_parts,
                                        FIELD_PARTS_DONE, num_parts_downloaded,
                                        FIELD_PARTS_FAILED, num_parts_failed,
                                        FIELD_PROGRESS, progress,
                                        FIELD_STATE_STR, state_str,
                                        FIELD_POSITION, position,
                                        FIELD_REAL_FILENAME, filename, -1);

    g_free(size_remaining);
    g_free(size);
    g_free(stamp_str);
}

static void
foreach_group_func (const char *collection_name, const char *subject, const char *group, gpointer data)
{
    GtkTreeStore *store = GTK_TREE_STORE(data);
    GtkTreeIter iter, parent;
    char *old_groups;
    char *new_groups;

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, FIELD_GROUPS, &old_groups, -1);
    if (old_groups && strlen(old_groups) > 0) {
        new_groups = g_strdup_printf("%s, %s", old_groups, group);
    } else {
        new_groups = g_strdup(group);
    }

    gtk_tree_store_set(store, &iter, FIELD_GROUPS, new_groups, -1);

    g_free(new_groups);
    g_free(old_groups);
}

static void
on_tree_view_select (GtkTreeSelection *selection, gpointer user_data)
{
    GtkWidget *btnRemoveItemFromQueue = nntpgrab_gui_base_get_widget("btnRemoveItemFromQueue");
    GtkWidget *btnQueueMoveToTop = nntpgrab_gui_base_get_widget("btnQueueMoveToTop");
    GtkWidget *btnQueueMoveUp = nntpgrab_gui_base_get_widget("btnQueueMoveUp");
    GtkWidget *btnQueueMoveDown = nntpgrab_gui_base_get_widget("btnQueueMoveDown");
    GtkWidget *btnQueueMoveToBottom = nntpgrab_gui_base_get_widget("btnQueueMoveToBottom");

    if (gtk_tree_selection_count_selected_rows(selection) == 0) {
        // No items selected, disable the remove button and the move buttons
        gtk_widget_set_sensitive(btnRemoveItemFromQueue, FALSE);
        gtk_widget_set_sensitive(btnQueueMoveToTop, FALSE);
        gtk_widget_set_sensitive(btnQueueMoveUp, FALSE);
        gtk_widget_set_sensitive(btnQueueMoveDown, FALSE);
        gtk_widget_set_sensitive(btnQueueMoveToBottom, FALSE);
        return;
    }

    // Enable the remove button and the move buttons
    gtk_widget_set_sensitive(btnRemoveItemFromQueue, TRUE);
    gtk_widget_set_sensitive(btnQueueMoveToTop, TRUE);
    gtk_widget_set_sensitive(btnQueueMoveUp, TRUE);
    gtk_widget_set_sensitive(btnQueueMoveDown, TRUE);
    gtk_widget_set_sensitive(btnQueueMoveToBottom, TRUE);
}

G_MODULE_EXPORT void
on_btnPauseQueue_clicked(GtkWidget *caller, gpointer data)
{
    SchedularState active_state = nntpgrab_schedular_get_state(core);

    if (status_flag == SCHEDULAR_STATE_STOPPED && active_state == SCHEDULAR_STATE_STOPPING) {
        GtkWidget *windowMain;

        windowMain = nntpgrab_gui_base_get_widget("windowMain");

        nntpgrab_gui_base_dialog_show(windowMain, _("The schedular is currently being stopped"), GTK_MESSAGE_INFO, GTK_BUTTONS_OK);

        g_signal_handlers_block_by_func (G_OBJECT (caller), G_CALLBACK (on_btnPauseQueue_clicked), data);

        gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(caller), TRUE);

        g_signal_handlers_unblock_by_func (G_OBJECT (caller), G_CALLBACK (on_btnPauseQueue_clicked), data);

        return;
    }

    if (status_flag == SCHEDULAR_STATE_RUNNING) {
        status_flag = SCHEDULAR_STATE_STOPPED;
        nntpgrab_schedular_stop(core, FALSE);
    } else {
        status_flag = SCHEDULAR_STATE_RUNNING;
        nntpgrab_schedular_start(core);
    }
}

static gboolean
on_key_press_event(GtkWidget *widget, GdkEventKey *key, gpointer data)
{
    if (key->keyval == GDK_Delete) {
        GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));
        if (gtk_tree_selection_count_selected_rows(selection) > 0) {
            on_btnRemoveItemFromQueue_clicked(widget, data);
            return TRUE;
        }
    }

    return FALSE;
}

static void
schedular_state_changed_cb(NNTPGrabCore *core, SchedularState state, const char *reason)
{
    GtkWidget *btnPauseQueue;

    if (state == SCHEDULAR_STATE_STOPPING) {
        status_flag = SCHEDULAR_STATE_STOPPED;
    } else {
        status_flag = state;
    }

    if (state == SCHEDULAR_STATE_STOPPING && reason != NULL) {
        char *msg;
        GtkWidget *windowMain;
#ifdef HAVE_LIBNOTIFY
        NotifyNotification *notification;
#endif

        windowMain = nntpgrab_gui_base_get_widget("windowMain");
        msg = g_strdup_printf(_("The schedular has been paused due to an error:\n%s"), reason);

#ifdef HAVE_LIBNOTIFY
        notification = notify_notification_new(_("Schedular has been paused"), msg, NULL, NULL);
        if (nntpgrab_gui_base_tray_icon_get_is_shown()) {
#if GTK_CHECK_VERSION(2,10,0)
            notify_notification_attach_to_status_icon(notification, nntpgrab_gui_base_tray_get_status_icon());
#endif
        }
        notify_notification_show(notification, NULL);
#endif

        nntpgrab_gui_base_dialog_show(windowMain, msg, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);
        g_free(msg);
    }

    btnPauseQueue = nntpgrab_gui_base_get_widget("btnPauseQueue");
    g_signal_handlers_block_by_func (G_OBJECT (btnPauseQueue), G_CALLBACK (on_btnPauseQueue_clicked), NULL);

    if (state == SCHEDULAR_STATE_RUNNING) {
        gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(btnPauseQueue), FALSE);
    } else {
        gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(btnPauseQueue), TRUE);
    }

    g_signal_handlers_unblock_by_func (G_OBJECT (btnPauseQueue), G_CALLBACK (on_btnPauseQueue_clicked), NULL);

}

static void
traffic_monitor_update_cb (NNTPGrabCore *core, int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, guint64 last_update_stamp, double average)
{
    static int times_been_here = 10;
    gboolean no_data_received = FALSE;
    GtkTreeModel *model;
    GtkWidget *treeDownloadQueue;
    GtkTreeIter iter, parent;
    guint64 total_estimated_time_remaining;

    if (bytes_received1 == 0 && bytes_received2 == 0 && bytes_received3 == 0 && bytes_received4 == 0 && bytes_received5 == 0 &&
        bytes_received6 == 0 && bytes_received7 == 0 && bytes_received8 == 0 && bytes_received9 == 0 && bytes_received10 == 0) {
            no_data_received = TRUE;
    }

    // Only update on startup, on every 3 seconds, and when there has no data been received in the last 10 seconds
    if (times_been_here < 3 && !no_data_received) {
        times_been_here += 1;
        return;
    }

    if (no_data_received) {
        times_been_here = 3;    // update the time remaining the next time we get here
    } else {
        times_been_here = 0;    // perform the next update in 10 seconds
    }

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");
    model = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    if (!gtk_tree_model_get_iter_first(model, &parent)) {
        // Nothing in the download queue
        return;
    }

    total_estimated_time_remaining = 0;

    do {
        guint64 filesize;
        int estimated_time_remaining;
        char *str_estimated_time_remaining;
        char *str_estimated_time_to_finish;

        gtk_tree_model_get(model, &parent, FIELD_FILESIZE_REMAINING, &filesize, -1);

        estimated_time_remaining = nntpgrab_utils_calculate_estimated_time_remaining(   bytes_received1, bytes_received2,
                                                                                        bytes_received3, bytes_received4,
                                                                                        bytes_received5, bytes_received6,
                                                                                        bytes_received7, bytes_received8,
                                                                                        bytes_received9, bytes_received10,
                                                                                        filesize);

        if (estimated_time_remaining == -1) {
            // Not enough data received to make an estimation
            return;
        }

        if (estimated_time_remaining == 0) {
            // File is already downloaded
            gtk_tree_store_set(GTK_TREE_STORE(model), &parent,  FIELD_ESTIMATED_TIME_REMAINING, "",
                                                                FIELD_ESTIMATED_TIME_TO_FINISH, "", -1);
        } else {
            str_estimated_time_remaining = nntpgrab_utils_get_readable_time_remaining(total_estimated_time_remaining + estimated_time_remaining);
            str_estimated_time_to_finish = nntpgrab_utils_get_readable_finished_time(total_estimated_time_remaining + estimated_time_remaining);

            gtk_tree_store_set(GTK_TREE_STORE(model), &parent,  FIELD_ESTIMATED_TIME_REMAINING, str_estimated_time_remaining,
                                                                FIELD_ESTIMATED_TIME_TO_FINISH, str_estimated_time_to_finish, -1);

            g_free(str_estimated_time_remaining);
            g_free(str_estimated_time_to_finish);
        }

        // Walk through all the children
        if (gtk_tree_model_iter_children(model, &iter, &parent)) {
            do {
                gtk_tree_model_get(model, &iter, FIELD_FILESIZE_REMAINING, &filesize, -1);

                if (filesize > 0) {
                    estimated_time_remaining = nntpgrab_utils_calculate_estimated_time_remaining(   bytes_received1, bytes_received2,
                                                                                                    bytes_received3, bytes_received4,
                                                                                                    bytes_received5, bytes_received6,
                                                                                                    bytes_received7, bytes_received8,
                                                                                                    bytes_received9, bytes_received10,
                                                                                                    filesize);

                    total_estimated_time_remaining += estimated_time_remaining;

                    str_estimated_time_remaining = nntpgrab_utils_get_readable_time_remaining(total_estimated_time_remaining);
                    str_estimated_time_to_finish = nntpgrab_utils_get_readable_finished_time(total_estimated_time_remaining);

                    gtk_tree_store_set(GTK_TREE_STORE(model), &iter,    FIELD_ESTIMATED_TIME_REMAINING, str_estimated_time_remaining,
                                                                        FIELD_ESTIMATED_TIME_TO_FINISH, str_estimated_time_to_finish, -1);

                    g_free(str_estimated_time_remaining);
                    g_free(str_estimated_time_to_finish);
                } else {
                    gtk_tree_store_set(GTK_TREE_STORE(model), &iter,    FIELD_ESTIMATED_TIME_REMAINING, "",
                                                                        FIELD_ESTIMATED_TIME_TO_FINISH, "", -1);
                }
            } while (gtk_tree_model_iter_next(model, &iter));
        }
    } while (gtk_tree_model_iter_next(model, &parent));
}

static void
config_changed(ConfigGUI *config, gpointer data)
{
    ConfigGUIOpts opts = config_gui_get_opts();

    gtk_tree_view_column_set_visible(columns[FIELD_PROGRESS], opts.show_column_progress);
    gtk_tree_view_column_set_visible(columns[FIELD_SUBJECT], opts.show_column_subject);
    gtk_tree_view_column_set_visible(columns[FIELD_POSTER], opts.show_column_poster);
    gtk_tree_view_column_set_visible(columns[FIELD_GROUPS], opts.show_column_newsgroups);
    gtk_tree_view_column_set_visible(columns[FIELD_FILESIZE_STR], opts.show_column_filesize);
    gtk_tree_view_column_set_visible(columns[FIELD_FILESIZE_REMAINING_STR], opts.show_column_filesize_remaining);
    gtk_tree_view_column_set_visible(columns[FIELD_STAMP_STR], opts.show_column_stamp);
    gtk_tree_view_column_set_visible(columns[FIELD_ESTIMATED_TIME_REMAINING], opts.show_column_time_remaining);
    gtk_tree_view_column_set_visible(columns[FIELD_ESTIMATED_TIME_TO_FINISH], opts.show_column_time_to_finish);
}

void
queue_initialize()
{
    GtkWidget *treeDownloadQueue;
    GtkTreeStore *store;
    GtkCellRenderer *cellRenderer;
    GtkTreeSelection *selection;

    treeDownloadQueue = nntpgrab_gui_base_get_widget("treeDownloadQueue");

    cellRenderer = gtk_cell_renderer_progress_new();
    columns[FIELD_PROGRESS] = gtk_tree_view_column_new_with_attributes(_("Progress"), cellRenderer, "value", FIELD_PROGRESS, "text", FIELD_STATE_STR, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_PROGRESS], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_PROGRESS], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_PROGRESS], -1);

    cellRenderer = gtk_cell_renderer_text_new();
    columns[FIELD_SUBJECT] = gtk_tree_view_column_new_with_attributes(_("Subject"), cellRenderer, "text", FIELD_SUBJECT, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_SUBJECT], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_SUBJECT], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_SUBJECT], -1);

    cellRenderer = gtk_cell_renderer_text_new();
    columns[FIELD_POSTER] = gtk_tree_view_column_new_with_attributes(_("Poster"), cellRenderer, "text", FIELD_POSTER, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_POSTER], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_POSTER], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_POSTER], -1);

    cellRenderer = gtk_cell_renderer_text_new();
    columns[FIELD_GROUPS] = gtk_tree_view_column_new_with_attributes(_("Group"), cellRenderer, "text", FIELD_GROUPS, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_GROUPS], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_GROUPS], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_GROUPS], -1);

    cellRenderer = gtk_cell_renderer_text_new();
    columns[FIELD_FILESIZE_STR] = gtk_tree_view_column_new_with_attributes(_("File size"), cellRenderer, "text", FIELD_FILESIZE_STR, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_FILESIZE_STR], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_FILESIZE_STR], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_FILESIZE_STR], -1);

    cellRenderer = gtk_cell_renderer_text_new();
    columns[FIELD_FILESIZE_REMAINING_STR] = gtk_tree_view_column_new_with_attributes(_("File size remaining"), cellRenderer, "text", FIELD_FILESIZE_REMAINING_STR, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_FILESIZE_REMAINING_STR], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_FILESIZE_REMAINING_STR], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_FILESIZE_REMAINING_STR], -1);

    cellRenderer = gtk_cell_renderer_text_new();
    columns[FIELD_STAMP_STR] = gtk_tree_view_column_new_with_attributes( _("Stamp"), cellRenderer, "text", FIELD_STAMP_STR, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_STAMP_STR], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_STAMP_STR], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_STAMP_STR], -1);

    cellRenderer = gtk_cell_renderer_text_new();
    columns[FIELD_ESTIMATED_TIME_REMAINING] = gtk_tree_view_column_new_with_attributes( _("Time remaining"), cellRenderer, "text", FIELD_ESTIMATED_TIME_REMAINING, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_ESTIMATED_TIME_REMAINING], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_ESTIMATED_TIME_REMAINING], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_ESTIMATED_TIME_REMAINING], -1);

    cellRenderer = gtk_cell_renderer_text_new();
    columns[FIELD_ESTIMATED_TIME_TO_FINISH] = gtk_tree_view_column_new_with_attributes( _("Time to finish"), cellRenderer, "text", FIELD_ESTIMATED_TIME_TO_FINISH, NULL);
    gtk_tree_view_column_set_resizable(columns[FIELD_ESTIMATED_TIME_TO_FINISH], TRUE);
    gtk_tree_view_column_set_reorderable(columns[FIELD_ESTIMATED_TIME_TO_FINISH], TRUE);
    gtk_tree_view_insert_column(GTK_TREE_VIEW(treeDownloadQueue), columns[FIELD_ESTIMATED_TIME_TO_FINISH], -1);

    store = gtk_tree_store_new( LAST_FIELD,
                                G_TYPE_STRING /* FIELD_COLLECTION_NAME */,
                                G_TYPE_STRING /* FIELD_SUBJECT */,
                                G_TYPE_STRING /* FIELD_POSTER */,
                                G_TYPE_STRING /* FIELD_GROUPS */,
                                G_TYPE_STRING /* FIELD_STAMP_STR */,
                                G_TYPE_INT /* FIELD_PROGRESS */,
                                G_TYPE_UINT64 /* FIELD_FILESIZE */,
                                G_TYPE_STRING /* FIELD_FILESIZE_STR */,
                                G_TYPE_UINT64 /* FIELD_FILESIZE_REMAINING */,
                                G_TYPE_STRING /* FIELD_FILESIZE_REMAINING_STR */,
                                G_TYPE_INT /* FIELD_NUM_PARTS */,
                                G_TYPE_INT /* FIELD_PARTS_DONE */,
                                G_TYPE_INT /* FIELD_PARTS_FAILED */,
                                G_TYPE_STRING /* FIELD_STATE_STR */,
                                G_TYPE_STRING /* FIELD_ESTIMATED_TIME_REMAINING */,
                                G_TYPE_STRING /* FIELD_ESTIMATED_TIME_TO_FINISH */,
                                G_TYPE_INT /* FIELD_POSITION */,
                                G_TYPE_STRING /* FIELD_REAL_FILENAME */);

    gtk_tree_view_set_model(GTK_TREE_VIEW(treeDownloadQueue), GTK_TREE_MODEL(store));

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));
    gtk_tree_selection_set_mode(GTK_TREE_SELECTION(selection), GTK_SELECTION_MULTIPLE);

    g_signal_connect(selection, "changed", G_CALLBACK(on_tree_view_select), NULL);
    on_tree_view_select(selection, NULL);

    nntpgrab_schedular_foreach_task(core, foreach_collection_func, foreach_file_func, foreach_group_func, store);

    g_signal_connect(core, "collection_added", G_CALLBACK(collection_added), NULL);
    g_signal_connect(core, "collection_removed", G_CALLBACK(collection_removed), NULL);
    g_signal_connect(core, "file_added", G_CALLBACK(file_added), NULL);
    g_signal_connect(core, "file_removed", G_CALLBACK(file_removed), NULL);
    g_signal_connect(core, "file_state_changed", G_CALLBACK(file_state_changed), NULL);
    g_signal_connect(core, "file_download_state_update", G_CALLBACK(file_download_state_update), NULL);
    g_signal_connect(core, "schedular_state_changed", G_CALLBACK(schedular_state_changed_cb), NULL);
    g_signal_connect(core, "traffic_monitor_update", G_CALLBACK(traffic_monitor_update_cb), NULL);
    g_signal_connect(core, "task_moved", G_CALLBACK(task_moved_cb), NULL);
    g_signal_connect(core, "collection_moved", G_CALLBACK(collection_moved_cb), NULL);

    g_signal_connect(config_gui_get_object(), "config_changed", G_CALLBACK(config_changed), NULL);
    config_changed(NULL, NULL);

    g_signal_connect(treeDownloadQueue, "key-press-event", G_CALLBACK(on_key_press_event), NULL);

#ifdef NNTPGRAB_NETWORKED
    status_flag = nntpgrab_schedular_get_state(core);
    if (status_flag == SCHEDULAR_STATE_STOPPING || status_flag == SCHEDULAR_STATE_STOPPED) {
        GtkWidget *btnPauseQueue = nntpgrab_gui_base_get_widget("btnPauseQueue");

        g_signal_handlers_block_by_func (G_OBJECT (btnPauseQueue), G_CALLBACK (on_btnPauseQueue_clicked), NULL);
        gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(btnPauseQueue), TRUE);
        g_signal_handlers_unblock_by_func (G_OBJECT (btnPauseQueue), G_CALLBACK (on_btnPauseQueue_clicked), NULL);
    }
#else
    nntpgrab_schedular_start(core);
#endif
}
