/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <glib-object.h>
#include "nntpgrab.h"
#include "collections.h"

#define IDLE_PRIORITY   G_PRIORITY_HIGH_IDLE + 50

NNTPGrabCore *get_core(void);

static gboolean
do_emit_fatal_error(gpointer data)
{
    char *errmsg = (char *) data;

    if (!get_core()) {
        g_free(errmsg);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "fatal_error", errmsg);
    g_free(errmsg);

    return FALSE;
}

void
nntpgrab_core_emit_fatal_error(const char *errmsg)
{
g_print("FATAL ERROR: %s\n", errmsg);

    if (!get_core()) {
        return;
    }

    g_idle_add_full(IDLE_PRIORITY, do_emit_fatal_error, (gpointer) g_strdup(errmsg), NULL);
}

static gboolean
do_emit_warning(gpointer data)
{
    char *errmsg = (char *) data;

    if (!get_core()) {
        g_free(errmsg);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "warning_message", errmsg);
    g_free(errmsg);

    return FALSE;
}

void
nntpgrab_core_emit_warning(const char *errmsg)
{
g_print("WARNING: %s\n", errmsg);

    if (!get_core()) {
        return;
    }

    g_idle_add_full(IDLE_PRIORITY, do_emit_warning, (gpointer) g_strdup(errmsg), NULL);
}

static gboolean
do_emit_config_changed(gpointer data)
{
    if (!get_core()) {
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "config_changed");

    return FALSE;
}

void
nntpgrab_core_emit_config_changed(void)
{
    if (!get_core()) {
        return;
    }

    g_idle_add_full(IDLE_PRIORITY, do_emit_config_changed, NULL, NULL);
}

struct _do_emit_connecting_data {
    char servername[128];
    int conn_id;
};

static gboolean
do_emit_connecting(gpointer data)
{
    struct _do_emit_connecting_data *connecting_data = (struct _do_emit_connecting_data *) data;

    if (!get_core()) {
        g_slice_free(struct _do_emit_connecting_data, connecting_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "connection_connecting", connecting_data->servername, connecting_data->conn_id);
    g_slice_free(struct _do_emit_connecting_data, connecting_data);

    return FALSE;
}

void
nntpgrab_core_emit_connecting(const char *servername, int conn_id)
{
    struct _do_emit_connecting_data *connecting_data;

    if (!get_core()) {
        return;
    }

    connecting_data = (struct _do_emit_connecting_data *) g_slice_new0(struct _do_emit_connecting_data);
    strncpy(connecting_data->servername, servername, sizeof(connecting_data->servername) - 1);
    connecting_data->conn_id = conn_id;

    g_idle_add_full(IDLE_PRIORITY, do_emit_connecting, connecting_data, NULL);
}

struct _do_emit_connected_data {
    char servername[128];
    int conn_id;
    char welcome_msg[1024];
};

static gboolean
do_emit_connected(gpointer data)
{
    struct _do_emit_connected_data *connected_data = (struct _do_emit_connected_data *) data;

    if (!get_core()) {
        g_slice_free(struct _do_emit_connected_data, connected_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "connection_connected", connected_data->servername, connected_data->conn_id, connected_data->welcome_msg);

    g_slice_free(struct _do_emit_connected_data, connected_data);

    return FALSE;
}

void
nntpgrab_core_emit_connected(const char *servername, int conn_id, const char *welcome_msg)
{
    struct _do_emit_connected_data *connected_data;

    if (!get_core()) {
        return;
    }

    connected_data = (struct _do_emit_connected_data *) g_slice_new0(struct _do_emit_connected_data);
    strncpy(connected_data->servername, servername, sizeof(connected_data->servername) - 1);
    connected_data->conn_id = conn_id;
    strncpy(connected_data->welcome_msg, welcome_msg, sizeof(connected_data->welcome_msg) - 1);

    g_idle_add_full(IDLE_PRIORITY, do_emit_connected, connected_data, NULL);
}

struct _do_emit_disconnect_data {
    char servername[128];
    int conn_id;
    NNTPDisconnectType disconnect_type;
    char errmsg[1024];
};

static gboolean
do_emit_disconnect(gpointer data)
{
    struct _do_emit_disconnect_data *disconnect_data = (struct _do_emit_disconnect_data *) data;

    if (!get_core()) {
        g_slice_free(struct _do_emit_disconnect_data, disconnect_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "connection_disconnect", disconnect_data->servername, disconnect_data->conn_id, disconnect_data->disconnect_type, strlen(disconnect_data->errmsg) > 0 ? disconnect_data->errmsg : NULL);

    g_slice_free(struct _do_emit_disconnect_data, disconnect_data);

    return FALSE;
}

void
nntpgrab_core_emit_disconnect(const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *errmsg)
{
    struct _do_emit_disconnect_data *disconnect_data;

    if (!get_core()) {
        return;
    }

    disconnect_data = g_slice_new0(struct _do_emit_disconnect_data);
    strncpy(disconnect_data->servername, servername, sizeof(disconnect_data->servername) - 1);
    disconnect_data->conn_id = conn_id;
    disconnect_data->disconnect_type = disconnect_type;
    if (errmsg) {
        strncpy(disconnect_data->errmsg, errmsg, sizeof(disconnect_data->errmsg) - 1);
    }

    g_idle_add_full(IDLE_PRIORITY, do_emit_disconnect, disconnect_data, NULL);
}

struct _traffic_data {
    int bytes_received[10];
    guint64 stamp;
};

static gboolean
do_emit_traffic_monitor_update(gpointer data)
{
    struct _traffic_data *traffic_data = (struct _traffic_data *) data;
    double average;
    int i;

    if (!get_core()) {
        g_slice_free(struct _traffic_data, traffic_data);
        return FALSE;
    }

    // Take the average from the last 5 seconds
    average = 0;
    for (i = 5; i < 10; i++) {
        average += traffic_data->bytes_received[i];
    }
    average /= 5;

    g_signal_emit_by_name(get_core(), "traffic_monitor_update", traffic_data->bytes_received[0],
                                                                traffic_data->bytes_received[1],
                                                                traffic_data->bytes_received[2],
                                                                traffic_data->bytes_received[3],
                                                                traffic_data->bytes_received[4],
                                                                traffic_data->bytes_received[5],
                                                                traffic_data->bytes_received[6],
                                                                traffic_data->bytes_received[7],
                                                                traffic_data->bytes_received[8],
                                                                traffic_data->bytes_received[9],
                                                                traffic_data->stamp,
                                                                average);

    g_slice_free(struct _traffic_data, traffic_data);

    return FALSE;
}

void
nntpgrab_core_emit_traffic_monitor_update(int bytes_received[10], time_t stamp)
{
    struct _traffic_data *traffic_data;

    if (!get_core()) {
        return;
    }

    traffic_data = (struct _traffic_data *) g_slice_new0(struct _traffic_data);
    memcpy(traffic_data->bytes_received, bytes_received, sizeof(int) * 10);
    traffic_data->stamp = stamp;

    g_idle_add_full(IDLE_PRIORITY, do_emit_traffic_monitor_update, traffic_data, NULL);
}

struct _part_data {
    char servername[128];
    int conn_id;
    char collection_name[256];
    char subject[256];
    int part_num;
    int size;
    gboolean all_servers_tried;
};

static gboolean
do_emit_part_download_start(gpointer data)
{
    struct _part_data *part_data = (struct _part_data *) data;

    if (!get_core()) {
        g_slice_free(struct _part_data, part_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "part_download_start", part_data->servername, part_data->conn_id, part_data->collection_name, part_data->subject, part_data->part_num);

    g_slice_free(struct _part_data, part_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_download_start(const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    struct _part_data *part_data;

    if (!get_core()) {
        return;
    }

    part_data = (struct _part_data *) g_slice_new0(struct _part_data);
    strncpy(part_data->servername, servername, sizeof(part_data->servername) - 1);
    part_data->conn_id = conn_id;
    strncpy(part_data->collection_name, collection_name, sizeof(part_data->collection_name) - 1);
    strncpy(part_data->subject, subject, sizeof(part_data->subject) - 1);
    part_data->part_num = part_num;

    g_idle_add_full(IDLE_PRIORITY, do_emit_part_download_start, part_data, NULL);
}

static gboolean
do_emit_part_failed(gpointer data)
{
    struct _part_data *part_data = (struct _part_data *) data;

    if (!get_core()) {
        g_slice_free(struct _part_data, part_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "part_failed", part_data->servername, part_data->conn_id, part_data->collection_name, part_data->subject, part_data->part_num, part_data->size, part_data->all_servers_tried);

    g_slice_free(struct _part_data, part_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_failed(const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gboolean all_servers_tried)
{
    struct _part_data *part_data;

    if (!get_core()) {
        return;
    }

    part_data = (struct _part_data *) g_slice_new0(struct _part_data);
    strncpy(part_data->servername, servername, sizeof(part_data->servername) - 1);
    part_data->conn_id = conn_id;
    strncpy(part_data->collection_name, collection_name, sizeof(part_data->collection_name) - 1);
    strncpy(part_data->subject, subject, sizeof(part_data->subject) - 1);
    part_data->part_num = part_num;
    part_data->all_servers_tried = all_servers_tried;
    part_data->size = size;

    g_idle_add_full(IDLE_PRIORITY, do_emit_part_failed, part_data, NULL);
}

static gboolean
do_emit_part_done(gpointer data)
{
    struct _part_data *part_data = (struct _part_data *) data;

    if (!get_core()) {
        g_slice_free(struct _part_data, part_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "part_done", part_data->servername, part_data->conn_id, part_data->collection_name, part_data->subject, part_data->part_num, part_data->size);

    g_slice_free(struct _part_data, part_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_done(const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size)
{
    struct _part_data *part_data;

    if (!get_core()) {
        return;
    }

    part_data = (struct _part_data *) g_slice_new0(struct _part_data);
    strncpy(part_data->servername, servername, sizeof(part_data->servername) - 1);
    part_data->conn_id = conn_id;
    strncpy(part_data->collection_name, collection_name, sizeof(part_data->collection_name) - 1);
    strncpy(part_data->subject, subject, sizeof(part_data->subject) - 1);
    part_data->part_num = part_num;
    part_data->size = size;

    g_idle_add_full(IDLE_PRIORITY, do_emit_part_done, part_data, NULL);
}

struct _part_progress_data {
    char servername[128];
    int conn_id;
    char subject[256];
    int part_num;
    int bytes_downloaded;
    int bytes_total;
};

static gboolean
do_emit_part_progress_update(gpointer data)
{
    struct _part_progress_data *part_data = (struct _part_progress_data *) data;

    if (!get_core()) {
        g_slice_free(struct _part_progress_data, part_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "part_progress_update", part_data->servername, part_data->conn_id, part_data->subject, part_data->part_num, part_data->bytes_downloaded, part_data->bytes_total);

    g_slice_free(struct _part_progress_data, part_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_progress_update(const char *servername, int conn_id, const char *subject, int part_num, int bytes_downloaded, int bytes_total)
{
    struct _part_progress_data *part_data;

    if (!get_core()) {
        return;
    }

    part_data = (struct _part_progress_data *) g_slice_new0(struct _part_progress_data);
    strncpy(part_data->servername, servername, sizeof(part_data->servername) - 1);
    part_data->conn_id = conn_id;
    strncpy(part_data->subject, subject, sizeof(part_data->subject) - 1);
    part_data->part_num = part_num;
    part_data->bytes_downloaded = bytes_downloaded;
    part_data->bytes_total = bytes_total;

    g_idle_add_full(IDLE_PRIORITY, do_emit_part_progress_update, part_data, NULL);
}

static gboolean
do_emit_collection_added(gpointer data)
{
    char *collection_name = (char *) data;

    if (!get_core()) {
        g_free(collection_name);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "collection_added", collection_name);

    g_free(collection_name);

    return FALSE;
}

void
nntpgrab_core_emit_collection_added(const char *collection_name)
{
    if (!get_core()) {
        return;
    }

    g_idle_add_full(IDLE_PRIORITY, do_emit_collection_added, g_strdup(collection_name), NULL);
}

static gboolean
do_emit_collection_removed(gpointer data)
{
    char *collection_name = (char *) data;

    if (!get_core()) {
        g_free(collection_name);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "collection_removed", collection_name);

    g_free(collection_name);

    return FALSE;
}

void
nntpgrab_core_emit_collection_removed(const char *collection_name)
{
    if (!get_core()) {
        return;
    }

    g_idle_add(do_emit_collection_removed, g_strdup(collection_name));
}

struct _file_added_data {
    char collection_name[256];
    char subject[256];
    char poster[256];
    guint64 stamp;
    guint64 file_size;
    guint64 total_file_size;
    guint64 total_file_size_remaining;
    int num_parts;
    TaskState status;
    GList *groups;
};

static gboolean
do_emit_file_added(gpointer data)
{
    struct _file_added_data *file_data = (struct _file_added_data *) data;
    GList *list;

    if (!get_core()) {
        list = file_data->groups;
        while (list) {
            g_free(list->data);
            list = g_list_next(list);
        }

        g_list_free(file_data->groups);
        g_slice_free(struct _file_added_data, file_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "file_added", file_data->collection_name, file_data->subject, file_data->poster, file_data->stamp, file_data->file_size, file_data->total_file_size, file_data->total_file_size_remaining, file_data->status, file_data->num_parts, file_data->groups);

    list = file_data->groups;
    while (list) {
        g_free(list->data);
        list = g_list_next(list);
    }

    g_list_free(file_data->groups);
    g_slice_free(struct _file_added_data, file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_added(const NNTPCollection *collection, const NNTPFile *file)
{
    struct _file_added_data *file_data;
    GList *list;

    if (!get_core()) {
        return;
    }

    file_data = (struct _file_added_data *) g_slice_new0(struct _file_added_data);
    strncpy(file_data->collection_name, collection->collection_name, sizeof(file_data->collection_name) - 1);
    strncpy(file_data->subject, file->subject, sizeof(file_data->subject) - 1);
    strncpy(file_data->poster, file->poster, sizeof(file_data->poster) - 1);
    file_data->stamp = file->stamp;
    file_data->file_size = file->file_size;
    file_data->status = file->status;
    file_data->num_parts = file->numparts;
    file_data->groups = NULL;
    file_data->total_file_size = collection->total_size;
    file_data->total_file_size_remaining = collection->total_size_remaining;

    list = file->groups;
    while (list) {
        char *group = (char *) list->data;
        file_data->groups = g_list_append(file_data->groups, g_strdup(group));
        list = g_list_next(list);
    }

    g_idle_add_full(IDLE_PRIORITY, do_emit_file_added, file_data, NULL);
}

struct _file_removed_data {
    char collection_name[256];
    char subject[256];
    guint64 total_file_size;
    guint64 total_file_size_remaining;
};

static gboolean
do_emit_file_removed(gpointer data)
{
    struct _file_removed_data *file_data = (struct _file_removed_data *) data;

    if (!get_core()) {
        g_slice_free(struct _file_removed_data, file_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "file_removed", file_data->collection_name, file_data->subject, file_data->total_file_size, file_data->total_file_size_remaining);

    g_slice_free(struct _file_removed_data, file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_removed(const char *collection_name, const char *subject, guint64 total_file_size, guint64 total_file_size_remaining)
{
    struct _file_removed_data *file_data;

    if (!get_core()) {
        return;
    }

    file_data = (struct _file_removed_data *) g_slice_new0(struct _file_removed_data);
    strncpy(file_data->collection_name, collection_name, sizeof(file_data->collection_name) - 1);
    strncpy(file_data->subject, subject, sizeof(file_data->subject) - 1);
    file_data->total_file_size = total_file_size;
    file_data->total_file_size_remaining = total_file_size_remaining;

    g_idle_add_full(IDLE_PRIORITY, do_emit_file_removed, file_data, NULL);
}

struct _file_download_state_update_data {
    char collection_name[256];
    char subject[256];
    int num_parts_total;
    int num_parts_done;
    int num_parts_failed;
    guint64 file_size;
    guint64 file_size_remaining;
    guint64 total_file_size;
    guint64 total_file_size_remaining;
};

static gboolean
do_emit_file_download_state_update(gpointer data)
{
    struct _file_download_state_update_data *file_data = (struct _file_download_state_update_data *) data;

    if (!get_core()) {
        g_slice_free(struct _file_download_state_update_data, file_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "file_download_state_update", file_data->collection_name, file_data->subject, file_data->num_parts_total, file_data->num_parts_done, file_data->num_parts_failed, file_data->file_size, file_data->file_size_remaining, file_data->total_file_size, file_data->total_file_size_remaining);

    g_slice_free(struct _file_download_state_update_data, file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_download_state_update(const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_file_size, guint64 total_file_size_remaining)
{
    struct _file_download_state_update_data *file_data;

    if (!get_core()) {
        return;
    }

    file_data = (struct _file_download_state_update_data *) g_slice_new0(struct _file_download_state_update_data);
    strncpy(file_data->collection_name, collection_name, sizeof(file_data->collection_name) - 1);
    strncpy(file_data->subject, subject, sizeof(file_data->subject) - 1);
    file_data->num_parts_total = num_parts_total;
    file_data->num_parts_done = num_parts_done;
    file_data->num_parts_failed = num_parts_failed;
    file_data->file_size = file_size;
    file_data->file_size_remaining = file_size_remaining;
    file_data->total_file_size = total_file_size;
    file_data->total_file_size_remaining = total_file_size_remaining;

    g_idle_add_full(IDLE_PRIORITY, do_emit_file_download_state_update, file_data, NULL);
}

struct _file_state_data {
    char collection_name[256];
    char subject[256];
    char real_filename[256];
    TaskState old_state;
    TaskState new_state;
};

static gboolean
do_emit_file_state_changed(gpointer data)
{
    struct _file_state_data *file_data = (struct _file_state_data *) data;
    char *real_filename;

    if (!get_core()) {
        g_slice_free(struct _file_state_data, file_data);
        return FALSE;
    }

    if (strlen(file_data->real_filename) > 0) {
        real_filename = file_data->real_filename;
    } else {
        real_filename = NULL;
    }

    g_signal_emit_by_name(get_core(), "file_state_changed", file_data->collection_name, file_data->subject, real_filename, file_data->old_state, file_data->new_state);

    g_slice_free(struct _file_state_data, file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_state_changed(const char *collection_name, const char *subject, const char *real_filename, TaskState old_state, TaskState new_state)
{
    struct _file_state_data *file_data;

    if (!get_core()) {
        return;
    }

    file_data = (struct _file_state_data *) g_slice_new0(struct _file_state_data);
    strncpy(file_data->collection_name, collection_name, sizeof(file_data->collection_name) - 1);
    strncpy(file_data->subject, subject, sizeof(file_data->subject) - 1);
    if (real_filename) {
        strncpy(file_data->real_filename, real_filename, sizeof(file_data->real_filename) - 1);
    }
    file_data->old_state = old_state;
    file_data->new_state = new_state;

    g_idle_add_full(IDLE_PRIORITY, do_emit_file_state_changed, file_data, NULL);
}

struct _schedular_data {
    SchedularState state;
    char reason[1024];
};

static gboolean
do_emit_schedular_state_changed(gpointer data)
{
    struct _schedular_data *schedular_data = (struct _schedular_data *) data;

    if (!get_core()) {
        g_slice_free(struct _schedular_data, schedular_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "schedular_state_changed", schedular_data->state, (strlen(schedular_data->reason) > 0) ? schedular_data->reason : NULL);

    g_slice_free(struct _schedular_data, schedular_data);

    return FALSE;
}

void
nntpgrab_core_emit_schedular_state_changed(SchedularState state, const char *reason)
{
    struct _schedular_data *schedular_data;

    if (!get_core()) {
        return;
    }

    schedular_data = (struct _schedular_data *) g_slice_new0(struct _schedular_data);
    schedular_data->state = state;
    if (reason) {
        strncpy(schedular_data->reason, reason, sizeof(schedular_data->reason) - 1);
    }

    g_idle_add_full(IDLE_PRIORITY, do_emit_schedular_state_changed, schedular_data, NULL);
}

struct _debug_data {
    GThread *thread;
    char message[1024];
};

static gboolean
do_emit_debug_message(gpointer data)
{
    struct _debug_data *debug_data = (struct _debug_data *) data;

    if (!get_core()) {
        g_slice_free(struct _debug_data, debug_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "debug_message", debug_data->thread, debug_data->message);

    g_slice_free(struct _debug_data, debug_data);

    return FALSE;
}

void
nntpgrab_core_emit_debug_message(const char *message)
{
    struct _debug_data *debug_data;

    if (!get_core()) {
        return;
    }

    debug_data = (struct _debug_data *) g_slice_new0(struct _debug_data);
    debug_data->thread = g_thread_self();
    strncpy(debug_data->message, message, sizeof(debug_data->message) - 1);

    g_idle_add_full(IDLE_PRIORITY, do_emit_debug_message, debug_data, NULL);
}

struct _move_data {
    char orig_collection_name[256];
    char new_collection_name[256];
    char subject[256];
    int old_position;
    int new_position;
};

static gboolean
do_emit_task_moved(gpointer data)
{
    struct _move_data *move_data = (struct _move_data *) data;

    if (!get_core()) {
        g_slice_free(struct _move_data, move_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "task_moved", move_data->orig_collection_name, move_data->subject, move_data->new_collection_name, move_data->old_position, move_data->new_position);

    g_slice_free(struct _move_data, move_data);

    return FALSE;
}

void
nntpgrab_core_emit_task_moved(const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position)
{
    struct _move_data *move_data;

    if (!get_core()) {
        return;
    }

    move_data = (struct _move_data *) g_slice_new0(struct _move_data);
    strncpy(move_data->orig_collection_name, orig_collection_name, sizeof(move_data->orig_collection_name) - 1);
    strncpy(move_data->subject, subject, sizeof(move_data->subject) - 1);
    strncpy(move_data->new_collection_name, new_collection_name, sizeof(move_data->new_collection_name) - 1);
    move_data->old_position = old_position;
    move_data->new_position = new_position;

    g_idle_add_full(IDLE_PRIORITY, do_emit_task_moved, move_data, NULL);
}

static gboolean
do_emit_collection_moved(gpointer data)
{
    struct _move_data *move_data = (struct _move_data *) data;

    if (!get_core()) {
        g_slice_free(struct _move_data, move_data);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "collection_moved", move_data->orig_collection_name, move_data->old_position, move_data->new_position);

    g_slice_free(struct _move_data, move_data);

    return FALSE;
}

void
nntpgrab_core_emit_collection_moved(const char *collection_name, int old_position, int new_position)
{
    struct _move_data *move_data;

    if (!get_core()) {
        return;
    }

    move_data = (struct _move_data *) g_slice_new0(struct _move_data);
    strncpy(move_data->orig_collection_name, collection_name, sizeof(move_data->orig_collection_name) - 1);
    move_data->old_position = old_position;
    move_data->new_position = new_position;

    g_idle_add_full(IDLE_PRIORITY, do_emit_collection_moved, move_data, NULL);
}

static gboolean
do_emit_par2_repair_message(gpointer data)
{
    char *msg = (char *) data;

    if (!get_core()) {
        g_free(msg);
        return FALSE;
    }

    g_signal_emit_by_name(get_core(), "par2_repair_message", msg);
    g_free(msg);

    return FALSE;
}

void
nntpgrab_core_emit_par2_repair_message(const char *msg)
{
    if (!get_core()) {
        return;
    }

    g_idle_add_full(IDLE_PRIORITY, do_emit_par2_repair_message, (gpointer) g_strdup(msg), NULL);
}
