/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>

#include "nntpgrab_plugin_par2.h"

static PluginPar2ImportedFuncs imported_funcs;

#ifdef WIN32
#include <windows.h>
#include <io.h> 

#define popen pt_popen
#define pclose pt_pclose

HANDLE my_pipein[2], my_pipeout[2], my_pipeerr[2];
char   my_popenmode = ' ';

static int
my_pipe(HANDLE *readwrite)
{
    SECURITY_ATTRIBUTES sa;

    sa.nLength = sizeof(sa);
    sa.bInheritHandle = 1;
    sa.lpSecurityDescriptor = NULL;

    if (!CreatePipe (&readwrite[0],&readwrite[1],&sa,1 << 13)) {
        errno = EMFILE;
        return -1;
    }

    return 0;
}

FILE *
pt_popen(const char *cmd, const char *mode)
{
    FILE *fptr = (FILE *)0;
    PROCESS_INFORMATION piProcInfo;
    STARTUPINFO siStartInfo;
    int success, catch_stderr;

    my_pipein[0]   = INVALID_HANDLE_VALUE;
    my_pipein[1]   = INVALID_HANDLE_VALUE;
    my_pipeout[0]  = INVALID_HANDLE_VALUE;
    my_pipeout[1]  = INVALID_HANDLE_VALUE;
    my_pipeerr[0]  = INVALID_HANDLE_VALUE;
    my_pipeerr[1]  = INVALID_HANDLE_VALUE;

    if (!mode || !*mode)
        goto finito;

    my_popenmode = *mode;
    if (my_popenmode != 'r' && my_popenmode != 'w')
        goto finito;

    catch_stderr = strstr((char *) cmd, "2>&1") != NULL;
    if (catch_stderr) {
        char *ptr = strstr((char*) cmd, "2>&1");
        *ptr = '\0';
    }

    if (my_pipe(my_pipein)  == -1 || my_pipe(my_pipeout) == -1)
        goto finito;
    if (!catch_stderr && my_pipe(my_pipeerr) == -1)
      goto finito;

    ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
    siStartInfo.cb             = sizeof(STARTUPINFO);
    siStartInfo.hStdInput      = my_pipein[0];
    siStartInfo.hStdOutput     = my_pipeout[1];
    if (catch_stderr)
        siStartInfo.hStdError  = my_pipeout[1];
    else
        siStartInfo.hStdError  = my_pipeerr[1];
    siStartInfo.dwFlags        = STARTF_USESTDHANDLES;

    success = CreateProcess(NULL,
       (LPTSTR)cmd,       // command line 
       NULL,              // process security attributes 
       NULL,              // primary thread security attributes 
       TRUE,              // handles are inherited 
       DETACHED_PROCESS,  // creation flags
       NULL,              // use parent's environment 
       NULL,              // use parent's current directory 
       &siStartInfo,      // STARTUPINFO pointer 
       &piProcInfo);      // receives PROCESS_INFORMATION 

    if (!success)
        goto finito;

    CloseHandle(my_pipein[0]);  my_pipein[0]  = INVALID_HANDLE_VALUE;
    CloseHandle(my_pipeout[1]); my_pipeout[1] = INVALID_HANDLE_VALUE;
    CloseHandle(my_pipeerr[1]); my_pipeerr[1] = INVALID_HANDLE_VALUE;

    if (my_popenmode == 'r')
        fptr = _fdopen(_open_osfhandle((long)my_pipeout[0],_O_BINARY),"r");
    else
        fptr = _fdopen(_open_osfhandle((long)my_pipein[1],_O_BINARY),"w");

finito:
    if (!fptr) {
        if (my_pipein[0]  != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipein[0]);
        if (my_pipein[1]  != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipein[1]);
        if (my_pipeout[0] != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipeout[0]);
        if (my_pipeout[1] != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipeout[1]);
        if (my_pipeerr[0] != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipeerr[0]);
        if (my_pipeerr[1] != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipeerr[1]);
    }

    return fptr;
}

int
pt_pclose(FILE *file)
{
    if (!file) {
        return -1;
    }

    fclose(file);

    CloseHandle(my_pipeerr[0]);
    if (my_popenmode == 'r')
        CloseHandle(my_pipein[1]);
    else
       CloseHandle(my_pipeout[0]);

    return 0;
}
#endif

gboolean
nntpgrab_plugin_par2_initialize(PluginPar2ImportedFuncs funcs, char **errmsg)
{
    imported_funcs = funcs;

    return TRUE;
}

void
nntpgrab_plugin_par2_destroy(void)
{
}

int
nntpgrab_plugin_par2_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

#if 0
static bool
LoadVerificationPacket(DiskFile *diskfile, u64 offset, PACKET_HEADER &header)
{
  VerificationPacket *packet = new VerificationPacket;

  // Load the packet from disk
  if (!packet->Load(diskfile, offset, header))
  {
    delete packet;
    return false;
  }

  // What is the fileid
  const MD5Hash &fileid = packet->FileId();

  // Look up the fileid in the source file map for an existing source file entry
  map<MD5Hash, Par2RepairerSourceFile*>::iterator sfmi = sourcefilemap.find(fileid);
  Par2RepairerSourceFile *sourcefile = (sfmi == sourcefilemap.end()) ? 0 :sfmi->second;

  // Was there an existing source file
  if (sourcefile)
  {
    // Does the source file already have a verification packet
    if (sourcefile->GetVerificationPacket())
    {
      // Yes. We don't need another copy.
      delete packet;
      return false;
    }
    else
    {
      // No. Store the packet in the source file
      sourcefile->SetVerificationPacket(packet);

      return true;
    }
  }
  else
  {
    // Create a new source file for the packet
    sourcefile = new Par2RepairerSourceFile(NULL, packet);

    // Record the source file in the source file map
    sourcefilemap.insert(pair<MD5Hash, Par2RepairerSourceFile*>(fileid, sourcefile));

    return true;
  }
}
#endif

gboolean
nntpgrab_plugin_par2_load_verification_packet(const char *filename, PAR2Set **par2set)
{
#if 0
    DiskFile *file;
    VerificationPacket *packet = NULL;

    g_return_val_if_fail(filename != NULL, FALSE);
    g_return_val_if_fail(par2set != NULL, FALSE);

    *par2set = NULL;

    file = new DiskFile();

    if (!file->Open(filename)) {
        delete file;
        return FALSE;
    }

    g_print("Now reading '%s' for PAR2 verification data..\n", filename);

    // How many useable packets have we found
    uint32_t packets = 0;

    // How big is the file
    guint64 filesize = file->FileSize();
    if (filesize <= 0) {
        delete file;
        g_print("Filesize == 0\n");
        return FALSE;
    }

    // Allocate a buffer to read data into
    // The buffer should be large enough to hold a whole
    // critical packet (i.e. file verification, file description, main,
    // and creator), but not necessarily a whole recovery packet.
    size_t buffersize = (size_t)min((u64)1048576, filesize);
    u8 *buffer = new u8[buffersize];

    // Start at the beginning of the file
    guint64 offset = 0;

    // Continue as long as there is at least enough for the packet header
    while (offset + sizeof(PACKET_HEADER) <= filesize) {
        // Attempt to read the next packet header
        PACKET_HEADER header;
        if (!file->Read(offset, &header, sizeof(header)))
            break;

        // Does this look like it might be a packet
        if (packet_magic != header.magic) {
            offset++;

            // Is there still enough for at least a whole packet header
            while (offset + sizeof(PACKET_HEADER) <= filesize) {
                // How much can we read into the buffer
                size_t want = (size_t)min((u64)buffersize, filesize-offset);

                // Fill the buffer
                if (!file->Read(offset, buffer, want)) {
                    offset = filesize;
                    break;
                }

                // Scan the buffer for the magic value
                u8 *current = buffer;
                u8 *limit = &buffer[want-sizeof(PACKET_HEADER)];
                while (current <= limit && packet_magic != ((PACKET_HEADER*)current)->magic) {
                    current++;
                }

                // What file offset did we reach
                offset += current-buffer;

                // Did we find the magic
                if (current <= limit) {
                    memcpy(&header, current, sizeof(header));
                    break;
                }
            }

            // Did we reach the end of the file
            if (offset + sizeof(PACKET_HEADER) > filesize)
            {
                break;
            }
        }

        // We have found the magic
        // Check the packet length
        if (sizeof(PACKET_HEADER) > header.length || // packet length is too small
            0 != (header.length & 3) ||              // packet length is not a multiple of 4
            filesize < offset + header.length) {     // packet would extend beyond the end of the file
                offset++;
                continue;
        }

        // Compute the MD5 Hash of the packet
        MD5Context context;
        context.Update(&header.setid, sizeof(header)-offsetof(PACKET_HEADER, setid));

        // How much more do I need to read to get the whole packet
        u64 current = offset+sizeof(PACKET_HEADER);
        u64 limit = offset+header.length;
        while (current < limit) {
            size_t want = (size_t)min((u64)buffersize, limit-current);

            if (!file->Read(current, buffer, want))
                break;

            context.Update(buffer, want);

            current += want;
        }

        // Did the whole packet get processed
        if (current < limit) {
            offset++;
            continue;
        }

        // Check the calculated packet hash against the value in the header
        MD5Hash hash;
        context.Final(hash);
        if (hash != header.hash) {
            offset++;
            continue;
        }

        // The Set ID is now validated, generate the PAR2Set structure
        *par2set = (PAR2Set*) g_slice_new0(PAR2Set);
        memcpy((*par2set)->set_id, header.setid.hash, sizeof(header.setid.hash));

        if (fileverificationpacket_type == header.type)  {
            //if (LoadVerificationPacket(file, offset, header)) {
                packets++;
            //}

            // Advance to the next packet
            offset += header.length;
        }
    }

    delete [] buffer;

    // We have finished with the file for now
    file->Close();

    // Did we actually find any interesting packets
    if (packets == 0) {
        if ((*par2set)) {
            g_slice_free(PAR2Set, (*par2set));
        }

        delete packet;
        delete file;
        return FALSE;
    }

    delete packet;
    delete file;
#endif

    return TRUE;
}

gboolean
nntpgrab_plugin_par2_repair_files(const char *directory, const char *par2filename)
{
    FILE *fp;
    char *cmd;
    char *path;
    char buf[1024];
    char *msg;
#ifdef WIN32
    char *dirname = g_locale_from_utf8(directory, -1, NULL, NULL, NULL);
    char *filename;
    char *bindir = g_path_get_dirname(__argv[0]);
    char *cmd2;

    g_return_val_if_fail(dirname != NULL, FALSE);
#endif

    path = g_build_filename(directory, par2filename, NULL);
    g_return_val_if_fail(g_file_test(path, G_FILE_TEST_EXISTS), FALSE);

    imported_funcs.emit_repair_message(_("Now starting PAR2 repair"));
#ifdef WIN32
    filename = g_locale_from_utf8(path, -1, NULL, NULL, NULL);

    g_return_val_if_fail(filename != NULL, FALSE);

    cmd = g_strdup_printf("\"%s\\par2.exe\" r -q \"%s\" \"%s"G_DIR_SEPARATOR_S"*.*\" 2>&1", bindir, filename, dirname);
    g_free(dirname);
    g_free(filename);
    g_free(bindir);
#else
    cmd = g_strdup_printf("par2 r -q \"%s\" \"%s"G_DIR_SEPARATOR_S"*\" 2>&1", path, directory);
#endif
    fp = popen(cmd, "r");
    g_free(path);

#ifdef WIN32
    cmd2 = g_locale_to_utf8(cmd, -1, NULL, NULL, NULL);
    g_return_val_if_fail(cmd2 != NULL, FALSE);
    msg = g_strdup_printf(_("Launched the command: %s\n"), cmd2);
    g_free(cmd2);
#else
    msg = g_strdup_printf(_("Launched the command: %s\n"), cmd);
#endif
    imported_funcs.emit_repair_message(msg);
    g_free(msg);
 
    if (!fp) {
        msg = g_strdup_printf(_("Error while launching from the par2cmdline process\n%s\n%s"), cmd, strerror(errno));
        imported_funcs.emit_repair_message(msg);
        g_free(cmd);
        g_free(msg);
        return FALSE;
    }
    g_free(cmd);

    memset(buf, 0, sizeof(buf));
    while (fgets(buf, sizeof(buf) - 1, fp)) {
        imported_funcs.emit_repair_message(buf);
        memset(buf, 0, sizeof(buf));
    }

    pclose(fp);

    imported_funcs.emit_repair_message("");
    imported_funcs.emit_repair_message(_("PAR2 repair finished"));
    imported_funcs.emit_repair_message("=====================================================");
    imported_funcs.emit_repair_message("");

    return TRUE;
}
