/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#include "nntpgrab.h"
#include "nntpgrab_utils.h"

#ifdef _MSC_VER
#undef atoll
#define atoll _atoi64
#endif

void
server_config_get_avail_servers(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    GList *servers, *list;

    servers = nntpgrab_config_get_avail_servers(core);
    list = servers;
    while (list) {
        const char *servername = (const char*) list->data;

        g_async_queue_push(send_buffer, g_strdup_printf("%s\n", servername));

        list = g_list_next(list);
    }

    g_async_queue_push(send_buffer, g_strdup("NULL\n"));

    nntpgrab_config_free_avail_servers(core, servers);
}

void
server_config_get_server_info(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    char servername[128];
    ConfigServer config;

    memset(servername, 0, sizeof(servername));

    if (sscanf(command, "CONFIG_GET_SERVER_INFO %127c", servername) != 1) {
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    if (!nntpgrab_config_get_server_info(core, servername, &config)) {
        g_async_queue_push(send_buffer, g_strdup("NULL\n"));
        return;
    }

    g_async_queue_push(send_buffer, g_strdup_printf("%s\n", config.servername));
    g_async_queue_push(send_buffer, g_strdup_printf("%s\n", config.hostname));
    g_async_queue_push(send_buffer, g_strdup_printf("%i\n", config.port));
    g_async_queue_push(send_buffer, g_strdup_printf("%s\n", config.username));
    g_async_queue_push(send_buffer, g_strdup_printf("%s\n", config.password));
    g_async_queue_push(send_buffer, g_strdup_printf("%i\n", config.max_threads));
    g_async_queue_push(send_buffer, g_strdup_printf("%i\n", config.priority));
    if (config.use_ssl) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }

    if (config.enabled) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }

    g_async_queue_push(send_buffer, g_strdup("NULL\n"));
}

void
server_config_add_server(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    ConfigServer server;
    NNTPGRAB_REGEX *re;
    const char **values;

    memset(&server, 0, sizeof(server));

    re = nntpgrab_utils_regex_compile("CONFIG_ADD_SERVER ([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]*)\t([^\t]*)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)");
    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7] && values[8]  && values[9])) {
        nntpgrab_utils_regex_matches_free(values);
        nntpgrab_utils_regex_free(re);
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    strncpy(server.servername, values[1], sizeof(server.servername));
    strncpy(server.hostname, values[2], sizeof(server.hostname));
    server.port = atoi(values[3]);
    strncpy(server.username, values[4], sizeof(server.username));
    strncpy(server.password, values[5], sizeof(server.password));
    server.max_threads = atoi(values[6]);
    server.priority = atoi(values[7]);

    if (!strncmp(values[8], "TRUE", 4)) {
        server.use_ssl = TRUE;
    } else {
        server.use_ssl = FALSE;
    }

    if (!strncmp(values[9], "TRUE", 4)) {
        server.enabled = TRUE;
    } else {
        server.enabled = FALSE;
    }

    nntpgrab_utils_regex_matches_free(values);
    nntpgrab_utils_regex_free(re);

    if (nntpgrab_config_add_server(core, server)) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }
}

void
server_config_del_server(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    char servername[128];

    memset(servername, 0, sizeof(servername));

    if (sscanf(command, "CONFIG_DEL_SERVER %127c", servername) != 1) {
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    if (nntpgrab_config_del_server(core, servername)) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }
}

void
server_config_edit_server(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    ConfigServer server;
    NNTPGRAB_REGEX *re;
    const char **values;

    memset(&server, 0, sizeof(server));

    re = nntpgrab_utils_regex_compile("CONFIG_EDIT_SERVER ([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]*)\t([^\t]*)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)");
    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7] && values[8] && values[9] && values[10])) {
        nntpgrab_utils_regex_matches_free(values);
        nntpgrab_utils_regex_free(re);
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    strncpy(server.servername, values[2], sizeof(server.servername));
    strncpy(server.hostname, values[3], sizeof(server.hostname));
    server.port = atoi(values[4]);
    strncpy(server.username, values[5], sizeof(server.username));
    strncpy(server.password, values[6], sizeof(server.password));
    server.max_threads = atoi(values[7]);
    server.priority = atoi(values[8]);

    if (!strncmp(values[9], "TRUE", 4)) {
        server.use_ssl = TRUE;
    } else {
        server.use_ssl = FALSE;
    }

    if (!strncmp(values[10], "TRUE", 4)) {
        server.enabled = TRUE;
    } else {
        server.enabled = FALSE;
    }

    if (nntpgrab_config_edit_server(core, values[1], server)) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }

    nntpgrab_utils_regex_matches_free(values);
    nntpgrab_utils_regex_free(re);
}

void
server_config_get_opts(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    ConfigOpts opts = nntpgrab_config_get_opts(core);

    g_async_queue_push(send_buffer, g_strdup_printf("%s\n", opts.download_directory));
    g_async_queue_push(send_buffer, g_strdup_printf("%s\n", opts.temp_directory));
    if (opts.enable_intelligent_par2_downloading) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }
    if (opts.enable_par2_repair) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }
    if (opts.enable_auto_import) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }
    g_async_queue_push(send_buffer, g_strdup_printf("%s\n", opts.auto_import_directory));
    if (opts.move_file_after_auto_import) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }
    g_async_queue_push(send_buffer, g_strdup("NULL\n"));
}

void
server_config_set_opts(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    ConfigOpts opts;
    NNTPGRAB_REGEX *re;
    const char **values;

    memset(&opts, 0, sizeof(opts));

    re = nntpgrab_utils_regex_compile("CONFIG_SET_OPTS ([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)");
    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7])) {
        nntpgrab_utils_regex_matches_free(values);
        nntpgrab_utils_regex_free(re);
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    strncpy(opts.download_directory, values[1], sizeof(opts.download_directory) - 1);
    strncpy(opts.temp_directory, values[2], sizeof(opts.temp_directory) - 1);

    if (!strncmp(values[3], "TRUE", 4)) {
        opts.enable_intelligent_par2_downloading = TRUE;
    } else {
        opts.enable_intelligent_par2_downloading = FALSE;
    }

    if (!strncmp(values[4], "TRUE", 4)) {
        opts.enable_par2_repair = TRUE;
    } else {
        opts.enable_par2_repair = FALSE;
    }

    if (!strncmp(values[5], "TRUE", 4)) {
        opts.enable_auto_import = TRUE;
    } else {
        opts.enable_auto_import = FALSE;
    }

    strncpy(opts.auto_import_directory, values[6], sizeof(opts.auto_import_directory) - 1);

    if (!strncmp(values[7], "TRUE", 4)) {
        opts.move_file_after_auto_import = TRUE;
    } else {
        opts.move_file_after_auto_import = FALSE;
    }

    nntpgrab_utils_regex_matches_free(values);
    nntpgrab_utils_regex_free(re);

    // Do the download directory, temp directory and nzb import directory exist?
    if (!g_file_test(opts.download_directory, G_FILE_TEST_IS_DIR | G_FILE_TEST_EXISTS) ||
        !g_file_test(opts.temp_directory, G_FILE_TEST_IS_DIR | G_FILE_TEST_EXISTS) ||
        !g_file_test(opts.auto_import_directory, G_FILE_TEST_IS_DIR | G_FILE_TEST_EXISTS)) {
            g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
            return;
    }

    nntpgrab_config_set_opts(core, opts);
    g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
}

void
server_config_get_folder_listing(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    GList *folders = NULL;
    GList *list;
    NNTPGRAB_REGEX *re;
    const char **values;

    re = nntpgrab_utils_regex_compile("CONFIG_GET_FOLDER_LISTING ([^\t]*)");
    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1])) {
        nntpgrab_utils_regex_matches_free(values);
        nntpgrab_utils_regex_free(re);
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    if (!nntpgrab_config_get_folder_listing(core, (values[1] && strlen(values[1]) > 0) ? values[1] : NULL, &folders)) {
        // Directory does not exist
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
        return;
    }

    nntpgrab_utils_regex_matches_free(values);
    nntpgrab_utils_regex_free(re);

    list = folders;
    while (list) {
        NNTPGrabFolder *folder = (NNTPGrabFolder*) list->data;
        char *msg;

        msg = g_strdup_printf("%s\t%s\n", folder->folder, (folder->has_subfolders ? "TRUE" : "FALSE"));

        g_async_queue_push(send_buffer, msg);

        list = g_list_next(list);
    }

    g_async_queue_push(send_buffer, g_strdup("NULL\n"));

    nntpgrab_config_free_folder_listing(core, folders);
}

void
server_schedular_start(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    if (!nntpgrab_schedular_start(core)) {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    }
}

void
server_schedular_stop(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    gboolean wait;
    char val[64];

    memset(val, 0, sizeof(val));

    if (sscanf(command, "SCHEDULAR_STOP %63c", val) != 1) {
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    if (!strncmp(val, "TRUE", 4)) {
        wait = TRUE;
    } else {
        wait = FALSE;
    }

    if (!nntpgrab_schedular_stop(core, wait)) {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    }
}

void
server_schedular_get_state(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    SchedularState state = nntpgrab_schedular_get_state(core);

    switch(state) {
        case SCHEDULAR_STATE_RUNNING:
            g_async_queue_push(send_buffer, g_strdup("RUNNING\n"));
            break;
        case SCHEDULAR_STATE_STOPPING:
            g_async_queue_push(send_buffer, g_strdup("STOPPING\n"));
            break;
        case SCHEDULAR_STATE_STOPPED:
            g_async_queue_push(send_buffer, g_strdup("STOPPED\n"));
            break;
    };
}

void
server_schedular_add_task_to_queue(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;
    GList *groups;
    GList *parts;
    GList *list;
    const char *collection_name;
    const char *subject;
    const char *poster;
    time_t stamp;
    guint64 file_size;
    char **values_groups;
    char **values_parts;
    int i;
    char *err = NULL;

    if (!re) {
        re = nntpgrab_utils_regex_compile("SCHEDULAR_ADD_TASK_TO_QUEUE ([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4] && values[5] && values[6] && values[7])) {
        nntpgrab_utils_regex_matches_free(values);
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    collection_name = values[1];
    subject = values[2];
    poster = values[3];
    stamp = atoi(values[4]);    // 64bit?
    file_size = atoll(values[5]);

    values_groups = g_strsplit(values[6], "|", -1);
    i = 0;
    groups = NULL;
    while (values_groups[i]) {
        groups = g_list_append(groups, values_groups[i]);
        i++;
    }

    values_parts = g_strsplit(values[7], "|", -1);
    i = 0;
    parts = NULL;
    while (values_parts[i]) {
        NNTPGrabPart *segment = g_slice_new(NNTPGrabPart);
        char **values_part = g_strsplit(values_parts[i], ",", 0);
        strncpy(segment->message_id,  values_part[0], sizeof(segment->message_id));
        segment->partnum = atoi(values_part[1]);
        segment->size = atoi(values_part[2]);
        g_strfreev(values_part);

        parts = g_list_append(parts, segment);
        i++;
    }

    if (nntpgrab_schedular_add_task_to_queue(core, collection_name, subject, poster, stamp, file_size, groups, parts, &err)) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        if (err) {
            g_async_queue_push(send_buffer, g_strdup_printf("FALSE %s\n", err));
            g_free(err);
        } else {
            g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
        }
    }

    list = parts;
    while (list) {
        g_slice_free(NNTPGrabPart, list->data);
        list = g_list_next(list);
    }

    g_list_free(groups);
    g_list_free(parts);
    g_strfreev(values_parts);
    g_strfreev(values_groups);
    nntpgrab_utils_regex_matches_free(values);
}

void
server_schedular_del_task_from_queue(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;
    char *err = NULL;

    if (!re) {
        re = nntpgrab_utils_regex_compile("SCHEDULAR_DEL_TASK_FROM_QUEUE ([^\t]+)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2])) {
        if (values) {
            nntpgrab_utils_regex_matches_free(values);
        }

        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    if (nntpgrab_schedular_del_task_from_queue(core, values[1], (strlen(values[2]) > 0) ? values[2] : NULL, &err)) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        if (err) {
            g_async_queue_push(send_buffer, g_strdup_printf("FALSE %s\n", err));
            g_free(err);
        } else {
            g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
        }
    }

    nntpgrab_utils_regex_matches_free(values);
}

void
server_schedular_restart_task(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;
    char *err = NULL;

    if (!re) {
        re = nntpgrab_utils_regex_compile("SCHEDULAR_RESTART_TASK ([^\t]+)\t([^\t]*)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2])) {
        nntpgrab_utils_regex_matches_free(values);
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    if (nntpgrab_schedular_restart_task(core, values[1], (strlen(values[2]) > 0) ? values[2] : NULL, &err)) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        if (err) {
            g_async_queue_push(send_buffer, g_strdup_printf("FALSE %s\n", err));
            g_free(err);
        } else {
            g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
        }
    }

    nntpgrab_utils_regex_matches_free(values);
}

void
server_schedular_save_queue(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    char *errmsg = NULL;

    if (!nntpgrab_schedular_save_queue(core, &errmsg)) {
        g_async_queue_push(send_buffer, g_strdup_printf("FALSE %s\n", errmsg));
        g_free(errmsg);
    } else {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    }
}

static gboolean group_needs_flush = FALSE;
static gboolean new_group = FALSE;

static void
foreach_collection_func (const char *collection_name, guint64 total_size, guint64 total_size_remaining, int position, gpointer data)
{
    GAsyncQueue *send_buffer = (GAsyncQueue *) data;
    char *line;

    if (group_needs_flush) {
        g_async_queue_push(send_buffer, g_strdup("\n"));
        group_needs_flush = FALSE;
    }

#if SIZEOF_LONG == 8
    line = g_strdup_printf("COLLECTION %s\t%li\t%li\t%i\n", collection_name, total_size, total_size_remaining, position);
#else
    line = g_strdup_printf("COLLECTION %s\t%lli\t%lli\t%i\n", collection_name, total_size, total_size_remaining, position);
#endif
    g_async_queue_push(send_buffer, line);
}

static void
foreach_file_func (const char *collection_name, const char *subject, const char *poster, gint64 stamp, guint64 file_size, guint64 file_size_remaining, int position, int num_parts_total, int num_parts_downloaded, int num_parts_failed, TaskState state, const char *filename, gpointer data)
{
    GAsyncQueue *send_buffer = (GAsyncQueue *) data;
    char *line;
    char *state_str;

    switch(state) {
        case TASK_STATE_WAITING_FOR_DOWNLOAD:
            state_str = "WAITING_FOR_DOWNLOAD";
            break;
        case TASK_STATE_DOWNLOADING:
            state_str = "DOWNLOADING";
            break;
        case TASK_STATE_WAITING_FOR_DECODE:
            state_str = "WAITING_FOR_DECODE";
            break;
        case TASK_STATE_DECODING:
            state_str = "DECODING";
            break;
        case TASK_STATE_FINISHED_COMPLETE:
            state_str = "FINISHED_COMPLETE";
            break;
        case TASK_STATE_FINISHED_INCOMPLETE:
            state_str = "FINISHED_INCOMPLETE";
            break;
        case TASK_STATE_FINISHED_NO_PARTS_AVAIL:
            state_str = "FINISHED_NO_PARTS_AVAIL";
            break;
        case TASK_STATE_SKIPPED:
            state_str = "SKIPPED";
            break;
        default:
            state_str = "FINISHED_COMPLETE";
    };

    if (group_needs_flush) {
        group_needs_flush = FALSE;
        g_async_queue_push(send_buffer, g_strdup("\n"));
    }

#if SIZEOF_LONG == 8
    line = g_strdup_printf("\tFILE %s\t%s\t%li\t%li\t%li\t%i\t%i\t%i\t%i\t%s\t%s\t", subject, poster, stamp, file_size, file_size_remaining, position, num_parts_total, num_parts_downloaded, num_parts_failed, state_str, filename);
#else
    line = g_strdup_printf("\tFILE %s\t%s\t%lli\t%lli\t%lli\t%i\t%i\t%i\t%i\t%s\t%s\t", subject, poster, stamp, file_size, file_size_remaining, position, num_parts_total, num_parts_downloaded, num_parts_failed, state_str, filename);
#endif
    g_async_queue_push(send_buffer, line);

    new_group = TRUE;
}

static void
foreach_group_func (const char *collection_name, const char *subject, const char *group, gpointer data)
{
    GAsyncQueue *send_buffer = (GAsyncQueue *) data;
    char *line;

    group_needs_flush = TRUE;

    if (new_group) {
        line = g_strdup_printf("%s", group);
        new_group = FALSE;
    } else {
        line = g_strdup_printf("|%s", group);
    }
    g_async_queue_push(send_buffer, line);
}

void
server_schedular_foreach_task(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    nntpgrab_schedular_foreach_task(core, foreach_collection_func, foreach_file_func, foreach_group_func, send_buffer);

    if (group_needs_flush) {
        group_needs_flush = FALSE;
        g_async_queue_push(send_buffer, g_strdup("\n"));
    }

    g_async_queue_push(send_buffer, g_strdup("NULL\n"));
}

void
server_schedular_move_task(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("SCHEDULAR_MOVE_TASK ([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2] && values[3] && values[4])) {
        nntpgrab_utils_regex_matches_free(values);
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    if (nntpgrab_schedular_move_task(core, values[1], (strlen(values[2]) > 0) ? values[2] : NULL, values[3], atoi(values[4]))) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }

    nntpgrab_utils_regex_matches_free(values);
}

void
server_schedular_move_collection(NNTPGrabCore *core, const char *command, GAsyncQueue *send_buffer)
{
    static NNTPGRAB_REGEX *re = NULL;
    const char **values;

    if (!re) {
        re = nntpgrab_utils_regex_compile("SCHEDULAR_MOVE_COLLECTION ([^\t]+)\t([^\t]+)");
    }

    values = nntpgrab_utils_regex_match(re, command);

    if (!(values && values[0] && values[1] && values[2])) {
        nntpgrab_utils_regex_matches_free(values);
        g_print(__FILE__ ":%i INVALID_COMMAND sent to client\n", __LINE__);
        g_async_queue_push(send_buffer, g_strdup("INVALID_COMMAND\n"));
        return;
    }

    if (nntpgrab_schedular_move_collection(core, values[1], atoi(values[2]))) {
        g_async_queue_push(send_buffer, g_strdup("TRUE\n"));
    } else {
        g_async_queue_push(send_buffer, g_strdup("FALSE\n"));
    }

    nntpgrab_utils_regex_matches_free(values);
}
